# WebGL Water
# https://madebyevan.com/webgl-water/
# Copyright 2011 Evan Wallace
# Released under the MIT license

import random
import math
import moderngl
import moderngl_window as mglw
from light_gl import *
from light_raytracer import *
from water import *
from renderer import *
from enum import IntEnum


class EventMode(IntEnum):
    NoEvent = 0,
    MoveCamera = 1,
    MoveSpahere = 2,
    AddDrop = 3,


class WaterMain(mglw.WindowConfig):
    """ coordinate
      | y
      |
      |       x
      +--------
     /
    /z
    """
    title = "ModernGL Water"
    gl_version = (3, 3)
    window_size = (1280, 720)
    aspect_ratio = 16 / 9
    resizable = True

    @property
    def angle_y(self):
        return self._angle_y

    @angle_y.setter
    def angle_y(self, value: float):
        self._angle_y = min(89.99, max(-89.99, value))

    @property
    def eye(self) -> glm.vec3:
        rad = glm.vec2(self.angle_x, self.angle_y) / 180 * math.pi
        cos_y = math.cos(rad[1])
        return glm.normalize(glm.vec3(math.cos(rad[0])*cos_y, math.sin(rad[1]), math.sin(rad[0])*cos_y)) * self.eye_distance

    def dbg_tracer_mesh(ctx: moderngl.Context, tracer: RayTracer, ray: glm.vec3) -> Mesh:
        a = tracer.eye + tracer.ray00 * 0.01
        b = tracer.eye + tracer.ray01 * 0.01
        c = tracer.eye + tracer.ray11 * 0.01
        d = tracer.eye + tracer.ray10 * 0.01
        return MeshBuilder.seg_lines(tracer.eye, a,
                                     tracer.eye, b,
                                     tracer.eye, c,
                                     tracer.eye, d,
                                     tracer.eye, ray,
                                     a, b, b, c, c, d, d, a,
                                     ).build(ctx)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.lines_shader = Shader("""
            void main() {
                gl_Position = ftransform();
            }
            """, """
            void main() {
                gl_FragColor = vec4(0.3, 0.7, 0.4, 1.0);
            }
            """, self.ctx)
        self.coord_y = MeshBuilder.seg_lines(
            glm.vec3(0.0, 0.0, 0.0), glm.vec3(0.0, 1.0, 0.0)).build(self.ctx)

        self.eye_distance = 3.0
        self.angle_x = 30
        self.angle_y = 30

        self.matrices = Matrices(glm.translate((0.0, 0.5, 0.0)), glm.lookAt(self.eye, (0.0, 0.0, 0.0), (0.0, 1.0, 0.0)), glm.perspective(
            45, self.aspect_ratio, 0.01, 100))
        self.renderer = Renderer(self.ctx)
        self.water = Water(ctx=self.ctx)
        self.center = glm.vec3(-0.4, -0.75, 0.2)
        self.old_center = self.center
        self.velocity = glm.vec3(0.0)
        self.gravity = glm.vec3(0.0, -4.0, 0.0)
        self.radius = 0.25
        self.use_gravity = False
        self.mode: EventMode = EventMode.NoEvent
        self.tracer: RayTracer = None
        self.tracer_mesh: Mesh = None

        for i in range(20):
            self.water.add_drop(random.random() * 2.0 - 1.0,
                                random.random() * 2.0 - 1.0, 0.03, 0.01 if i % 1 else -0.01)

    def add_drop(self, x: float, y: float):
        ray = self.tracer.get_ray_for_pixel(x, y)
        point_on_panel = self.tracer.eye + (ray * (-self.tracer.eye.y / ray.y))
        self.water.add_drop(point_on_panel.x, point_on_panel.z, 0.03, 0.01)

    def move_sphere(self, x: int, y: int):
        ray = self.tracer.get_ray_for_pixel(x, y)
        t = glm.dot(self.plane_normal, self.prev_hit -
                    self.tracer.eye) / glm.dot(self.plane_normal, ray)
        next_hit = self.tracer.eye + (ray * t)
        center = self.center + (next_hit - self.prev_hit)
        center.x = max(self.radius - 1, min(1 - self.radius, center.x))
        center.z = max(self.radius - 1, min(1 - self.radius, center.z))
        center.y = max(self.radius - 1, min(10, center.y))
        self.center = center
        self.prev_hit = next_hit

    def mouse_release_event(self, x: int, y: int, button: int):
        if button == 1:
            self.mode = EventMode.NoEvent
            self.tracer = None

    def mouse_press_event(self, x: int, y: int, button: int):
        self.tracer = RayTracer(self.ctx.viewport, self.matrices)
        ray = self.tracer.get_ray_for_pixel(x, y)
        sphere_hit_test: HitTest = RayTracer.hit_test_sphere(
            self.tracer.eye, ray, self.center, self.radius)
        point_on_panel = self.tracer.eye + (ray * (-self.tracer.eye.y / ray.y))
        if button == 1:
            if sphere_hit_test is not None:
                self.mode = EventMode.MoveSpahere
                self.prev_hit = sphere_hit_test.hit
                self.plane_normal = -self.tracer.get_ray_for_pixel(
                    self.ctx.viewport[2] / 2, self.ctx.viewport[3] / 2)
            elif abs(point_on_panel.x) < 1 and abs(point_on_panel.z) < 1:
                self.mode = EventMode.AddDrop
                self.add_drop(x, y)
            else:
                self.mode = EventMode.MoveCamera
        else:
            self.tracer_mesh = WaterMain.dbg_tracer_mesh(
                self.ctx, self.tracer, point_on_panel)

    def mouse_drag_event(self, x: int, y: int, dx: int, dy: int):
        match self.mode:
            case EventMode.AddDrop:
                self.add_drop(x, y)
            case EventMode.MoveCamera:
                self.angle_x += dx
                self.angle_y += dy
            case EventMode.MoveSpahere:
                self.move_sphere(x, y)

    def key_event(self, key, action, modifiers):
        if key == ord('G') and action == 1:
            self.use_gravity = not self.use_gravity

    def update(self, seconds):
        if seconds > 1:
            return
        if self.mode == EventMode.MoveSpahere:
            self.velocity = glm.vec3(0.0)
        elif self.use_gravity:
            precent_under_water = max(
                0, min(1, (self.radius - self.center.y) / (2*self.radius)))
            self.velocity = self.velocity + \
                (self.gravity * (seconds - 1.1 * seconds * precent_under_water))
            self.velocity = self.velocity - \
                (glm.normalize(self.velocity) * (precent_under_water *
                                                 seconds * glm.dot(self.velocity, self.velocity)))
            self.center = self.center + self.velocity * seconds
            if self.center.y < self.radius - 1:
                self.center.y = self.radius - 1
                self.velocity.y = abs(self.velocity.y) * 0.7

        self.water.move_sphere(self.old_center, self.center, self.radius)
        self.old_center = self.center

        self.water.step_simulation()
        self.water.step_simulation()
        self.water.update_normals()
        self.renderer.update_caustics(self.matrices, self.water)

    def render(self, time, frame_time):
        self.matrices.view = glm.lookAt(
            self.eye, (0.0, 0.0, 0.0), (0.0, 1.0, 0.0))

        self.update(frame_time)
        self.ctx.enable(moderngl.DEPTH_TEST)
        self.ctx.screen.use()
        self.renderer.sphere_center = self.center
        self.renderer.sphere_radius = self.radius
        self.renderer.render_cude(self.matrices, self.water)
        self.renderer.render_water(self.matrices, self.water)
        self.renderer.render_sphere(self.matrices, self.water)
        self.ctx.disable(moderngl.DEPTH_TEST)
        if self.tracer_mesh is not None:
            self.lines_shader.draw_mesh(
                self.tracer_mesh, self.matrices, mode=moderngl.LINES)


mglw.run_window_config(WaterMain, args=["--window", "glfw"])



import math
import os
import sys

import imgui
import moderngl
import pygame
from zengl_imgui import PygameBackend

os.environ['SDL_WINDOWS_DPI_AWARENESS'] = 'permonitorv2'

pygame.init()
pygame.display.set_mode((1280, 720), flags=pygame.OPENGL | pygame.DOUBLEBUF, vsync=True)


class MGLBackend(PygameBackend):
    def __init__(self):
        super().__init__()
        self.renderer.ctx.end_frame(flush=False)

    def render(self):
        self.renderer.ctx.new_frame(clear=False)
        super().render()
        self.renderer.ctx.end_frame()


impl = MGLBackend()


class Scene:
    def __init__(self):
        self.ctx = moderngl.get_context()

    def render(self):
        now = pygame.time.get_ticks() / 1000.0

        r = math.sin(now + 0.0) * 0.5 + 0.5
        g = math.sin(now + 2.1) * 0.5 + 0.5
        b = math.sin(now + 4.2) * 0.5 + 0.5
        self.ctx.clear(r, g, b)


scene = Scene()

while True:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            pygame.quit()
            sys.exit()
        impl.process_event(event)
    impl.process_inputs()

    imgui.new_frame()
    imgui.show_test_window()
    imgui.end_frame()
    imgui.render()

    scene.render()
    impl.render()

    pygame.display.flip()



from contextlib import ExitStack

import pytest
from OpenGL import GL

import moderngl

def test_debug_scope(ctx):
    with ctx.debug_scope("test scope"):
        pass


def test_debug_scope_error_when_label_is_too_long(ctx):
    max_label_length = ctx.max_debug_message_length
    message = "F" * (max_label_length + 1)

    with pytest.raises(moderngl.Error):
        with ctx.debug_scope(message):
            pass


def test_debug_scope_error_when_stack_overflow(ctx_new):
    max_stack_depth = ctx_new.max_debug_group_stack_depth
    with ExitStack() as stack:
        with pytest.raises(moderngl.Error):
            for _ in range(max_stack_depth + 1):
                stack.enter_context(ctx_new.debug_scope("test scope"))


def test_debug_scope_error_when_stack_underflow(ctx_new):
    if GL.glPopDebugGroup:
        with pytest.raises(moderngl.Error):
            with ctx_new.debug_scope("test scope"):
                GL.glPopDebugGroup()


def test_debug_scope_error_when_label_wrong_type(ctx):
    with pytest.raises(TypeError):
        with ctx.debug_scope(46):
            pass


import moderngl
import pytest

def test_error_when_label_is_too_long(ctx):
    buf = ctx.buffer(reserve=1024)
    max_label_length = ctx.max_label_length

    with pytest.raises(moderngl.Error):
        buf.label = "F" * (max_label_length + 1)


def test_error_with_wrong_type(ctx):
    buf = ctx.buffer(reserve=1024)

    with pytest.raises(TypeError):
        buf.label = 123


def test_no_label_on_creation(ctx):
    buf = ctx.buffer(reserve=1024)

    assert buf.label is None


def test_clearing_label(ctx):
    buf = ctx.buffer(reserve=1024)
    buf.label = "test buffer"
    assert buf.label is not None

    buf.label = None

    assert buf.label is None



# WebGL Water
# https://madebyevan.com/webgl-water/
# Copyright 2011 Evan Wallace
# Released under the MIT license

import os
import glm
from light_gl import *
from water import Water
import moderngl


def local(*path):
    return os.path.join(os.path.dirname(__file__), *path)

def local_textures(path):
    return local("../data/textures", path)

class Renderer():
    _helper_functions = """
    const float IOR_AIR = 1.0;
    const float IOR_WATER = 1.333;
    const vec3 abovewaterColor = vec3(0.25, 1.0, 1.25);
    const vec3 underwaterColor = vec3(0.4, 0.9, 1.0);
    const float poolHeight = 1.0;
    uniform vec3 light;
    uniform vec3 sphereCenter;
    uniform float sphereRadius;
    uniform sampler2D tiles;
    uniform sampler2D causticTex;
    uniform sampler2D water;
    
    vec2 intersectCube(vec3 origin, vec3 ray, vec3 cubeMin, vec3 cubeMax) {
        vec3 tMin = (cubeMin - origin) / ray;
        vec3 tMax = (cubeMax - origin) / ray;
        vec3 t1 = min(tMin, tMax);
        vec3 t2 = max(tMin, tMax);
        float tNear = max(max(t1.x, t1.y), t1.z);
        float tFar = min(min(t2.x, t2.y), t2.z);
        return vec2(tNear, tFar);
    }
    
    float intersectSphere(vec3 origin, vec3 ray, vec3 sphereCenter, float sphereRadius) {
        vec3 toSphere = origin - sphereCenter;
        float a = dot(ray, ray);
        float b = 2.0 * dot(toSphere, ray);
        float c = dot(toSphere, toSphere) - sphereRadius * sphereRadius;
        float discriminant = b*b - 4.0*a*c;
        if (discriminant > 0.0) {
        float t = (-b - sqrt(discriminant)) / (2.0 * a);
        if (t > 0.0) return t;
        }
        return 1.0e6;
    }
    
    vec3 getSphereColor(vec3 point) {
        vec3 color = vec3(0.5);
        
        /* ambient occlusion with walls */
        color *= 1.0 - 0.9 / pow((1.0 + sphereRadius - abs(point.x)) / sphereRadius, 3.0);
        color *= 1.0 - 0.9 / pow((1.0 + sphereRadius - abs(point.z)) / sphereRadius, 3.0);
        color *= 1.0 - 0.9 / pow((point.y + 1.0 + sphereRadius) / sphereRadius, 3.0);
        
        /* caustics */
        vec3 sphereNormal = (point - sphereCenter) / sphereRadius;
        vec3 refractedLight = refract(-light, vec3(0.0, 1.0, 0.0), IOR_AIR / IOR_WATER);
        float diffuse = max(0.0, dot(-refractedLight, sphereNormal)) * 0.5;
        vec4 info = texture2D(water, point.xz * 0.5 + 0.5);
        if (point.y < info.r) {
        vec4 caustic = texture2D(causticTex, 0.75 * (point.xz - point.y * refractedLight.xz / refractedLight.y) * 0.5 + 0.5);
        diffuse *= caustic.r * 4.0;
        }
        color += diffuse;
        
        return color;
    }
    
    vec3 getWallColor(vec3 point) {
        float scale = 0.5;
        
        vec3 wallColor;
        vec3 normal;
        if (abs(point.x) > 0.999) {
            wallColor = texture2D(tiles, point.yz * 0.5 + vec2(1.0, 0.5)).rgb;
            normal = vec3(-point.x, 0.0, 0.0);
        } else if (abs(point.z) > 0.999) {
            wallColor = texture2D(tiles, point.yx * 0.5 + vec2(1.0, 0.5)).rgb;
            normal = vec3(0.0, 0.0, -point.z);
        } else {
            wallColor = texture2D(tiles, point.xz * 0.5 + 0.5).rgb;
            normal = vec3(0.0, 1.0, 0.0);
        }
        
        scale /= length(point); /* pool ambient occlusion */
        scale *= 1.0 - 0.9 / pow(length(point - sphereCenter) / sphereRadius, 4.0); /* sphere ambient occlusion */
        
        /* caustics */
        vec3 refractedLight = -refract(-light, vec3(0.0, 1.0, 0.0), IOR_AIR / IOR_WATER);
        float diffuse = max(0.0, dot(refractedLight, normal));
        vec4 info = texture2D(water, point.xz * 0.5 + 0.5);
        if (point.y < info.r) {
            vec4 caustic = texture2D(causticTex, 0.75 * (point.xz - point.y * refractedLight.xz / refractedLight.y) * 0.5 + 0.5);
            scale += diffuse * caustic.r * 2.0 * caustic.g;
        } else {
            /* shadow for the rim of the pool */
            vec2 t = intersectCube(point, refractedLight, vec3(-1.0, -poolHeight, -1.0), vec3(1.0, 2.0, 1.0));
            diffuse *= 1.0 / (1.0 + exp(-200.0 / (1.0 + 10.0 * (t.y - t.x)) * (point.y + refractedLight.y * t.y - 2.0 / 12.0)));
            
            scale += diffuse * 0.5;
        }
        
        return wallColor * scale;
    }
    """

    def _splice(array: list, start: int, num: int):
        new_array = []
        exclude_range = range(start, start + num)
        for idx, elem in enumerate(array):
            if idx not in exclude_range:
                new_array.append(elem)
        return new_array

    def _water_shader(ctx: moderngl.Context, under_water: bool):
        vert_shader = """
        uniform sampler2D water;
        varying vec3 position;
        void main() {
            vec4 info = texture2D(water, gl_Vertex.xy * 0.5 + 0.5);
            position = gl_Vertex.xzy;
            position.y += info.r;
            gl_Position = gl_ModelViewProjectionMatrix * vec4(position, 1.0);
        }
        """
        if under_water:
            water_frag_shader = """
            /* underwater */ 
            normal = -normal;
            vec3 reflectedRay = reflect(incomingRay, normal);
            vec3 refractedRay = refract(incomingRay, normal, IOR_WATER / IOR_AIR);
            float fresnel = mix(0.5, 1.0, pow(1.0 - dot(normal, -incomingRay), 3.0));
            
            vec3 reflectedColor = getSurfaceRayColor(position, reflectedRay, underwaterColor);
            vec3 refractedColor = getSurfaceRayColor(position, refractedRay, vec3(1.0)) * vec3(0.8, 1.0, 1.1);
            
            gl_FragColor = vec4(mix(reflectedColor, refractedColor, (1.0 - fresnel) * length(refractedRay)), 1.0);
            """
        else:
            water_frag_shader = """
                /* above water */
                vec3 reflectedRay = reflect(incomingRay, normal);
                vec3 refractedRay = refract(incomingRay, normal, IOR_AIR / IOR_WATER);
                float fresnel = mix(0.25, 1.0, pow(1.0 - dot(normal, -incomingRay), 3.0));
                
                vec3 reflectedColor = getSurfaceRayColor(position, reflectedRay, abovewaterColor);
                vec3 refractedColor = getSurfaceRayColor(position, refractedRay, abovewaterColor);
                
                gl_FragColor = vec4(mix(refractedColor, reflectedColor, fresnel), 1.0);
            """

        frag_shader = """
        uniform vec3 eye;
        varying vec3 position;
        uniform samplerCube sky;
        
        vec3 getSurfaceRayColor(vec3 origin, vec3 ray, vec3 waterColor) {
            vec3 color;
                float q = intersectSphere(origin, ray, sphereCenter, sphereRadius);
            if (q < 1.0e6) {
                color = getSphereColor(origin + ray * q);
            } else if (ray.y < 0.0) {
                vec2 t = intersectCube(origin, ray, vec3(-1.0, -poolHeight, -1.0), vec3(1.0, 2.0, 1.0));
                color = getWallColor(origin + ray * t.y);
            } else {
                vec2 t = intersectCube(origin, ray, vec3(-1.0, -poolHeight, -1.0), vec3(1.0, 2.0, 1.0));
                vec3 hit = origin + ray * t.y;
                if (hit.y < 2.0 / 12.0) {
                    color = getWallColor(hit);
                } else {
                    color = textureCube(sky, ray).rgb;
                    color += vec3(pow(max(0.0, dot(light, ray)), 5000.0)) * vec3(10.0, 8.0, 6.0);
                }
            }
            if (ray.y < 0.0) color *= waterColor;
            return color;
        }
        
        void main() {
            vec2 coord = position.xz * 0.5 + 0.5;
            vec4 info = texture2D(water, coord);
            
            /* make water look more "peaked" */
            for (int i = 0; i < 5; i++) {
            coord += info.ba * 0.005;
            info = texture2D(water, coord);
            }
            
            vec3 normal = vec3(info.b, sqrt(1.0 - dot(info.ba, info.ba)), info.a);
            vec3 incomingRay = normalize(position - eye);
        """ + water_frag_shader + """}"""
        return Shader(vert_shader, Renderer._helper_functions + frag_shader, ctx)

    def __init__(self, ctx: moderngl.Context) -> None:
        self.ctx = ctx
        self.tile_texture = ImageTexture(
            local_textures("tiles.jpg"), "RGB", ctx, (True, True, True), mipmaps=True)
        self.light_dir = glm.normalize(glm.vec3(2.0, 2.0, -1.0))
        # water
        self.water_mesh = MeshBuilder.panel(detail=200).build(ctx)
        self.water_shaders: list[Shader] = []
        for under_water in (False, True):
            self.water_shaders.append(Renderer._water_shader(ctx, under_water))
        # sphere
        self.sphere_mesh = MeshBuilder.sphere(detail=10).build(ctx)
        self.sphere_shader = Shader(Renderer._helper_functions + """
        varying vec3 position;
        void main() {
            position = sphereCenter + gl_Vertex.xyz * sphereRadius;
            gl_Position = gl_ModelViewProjectionMatrix * vec4(position, 1.0);
        }
        """, Renderer._helper_functions + """
        varying vec3 position;
        void main() {
            gl_FragColor = vec4(getSphereColor(position), 1.0);
            vec4 info = texture2D(water, position.xz * 0.5 + 0.5);
            if (position.y < info.r) {
                gl_FragColor.rgb *= underwaterColor * 1.2;
            }
        }""", ctx)
        self.sphere_center = glm.vec3(0.0)
        self.sphere_radius = 0
        # cude
        cude_mesh_builder = MeshBuilder.cube()
        # remove cude top
        cude_mesh_builder.triangles = np.array(Renderer._splice(
            cude_mesh_builder.triangles, 4, 2), dtype=np.int32)
        self.cube_mesh = cude_mesh_builder.build(ctx)
        self.cude_shader = Shader(Renderer._helper_functions + """
        varying vec3 position;
        void main() {
            position = gl_Vertex.xyz;
            position.y = ((1.0 - position.y) * (7.0 / 12.0) - 1.0) * poolHeight;
            gl_Position = gl_ModelViewProjectionMatrix * vec4(position, 1.0);
        }""", Renderer._helper_functions + """
        varying vec3 position;
        void main() {
            gl_FragColor = vec4(getWallColor(position), 1.0);
            vec4 info = texture2D(water, position.xz * 0.5 + 0.5);
            if (position.y < info.r) {
                gl_FragColor.rgb *= underwaterColor * 1.2;
            }
        }""", ctx)

        self.caustic_texture = RawTexture((1024, 1024), ctx=ctx)
        self.caustics_shader = Shader(Renderer._helper_functions + """
        varying vec3 oldPos;
        varying vec3 newPos;
        varying vec3 ray;
        
        /* project the ray onto the plane */
        vec3 project(vec3 origin, vec3 ray, vec3 refractedLight) {
            vec2 tcube = intersectCube(origin, ray, vec3(-1.0, -poolHeight, -1.0), vec3(1.0, 2.0, 1.0));
            origin += ray * tcube.y;
            float tplane = (-origin.y - 1.0) / refractedLight.y;
            return origin + refractedLight * tplane;
        }
        
        void main() {
            vec4 info = texture2D(water, gl_Vertex.xy * 0.5 + 0.5);
            info.ba *= 0.5;
            vec3 normal = vec3(info.b, sqrt(1.0 - dot(info.ba, info.ba)), info.a);
            
            /* project the vertices along the refracted vertex ray */
            vec3 refractedLight = refract(-light, vec3(0.0, 1.0, 0.0), IOR_AIR / IOR_WATER);
            ray = refract(-light, normal, IOR_AIR / IOR_WATER);
            oldPos = project(gl_Vertex.xzy, refractedLight, refractedLight);
            newPos = project(gl_Vertex.xzy + vec3(0.0, info.r, 0.0), ray, refractedLight);

            gl_Position = vec4(0.75 * (newPos.xz + refractedLight.xz / refractedLight.y), 0.0, 1.0);
        }""", Renderer._helper_functions + """
        varying vec3 oldPos;
        varying vec3 newPos;
        varying vec3 ray;
        void main() {
            /* if the triangle gets smaller, it gets brighter, and vice versa */
            float oldArea = length(dFdx(oldPos)) * length(dFdy(oldPos));
            float newArea = length(dFdx(newPos)) * length(dFdy(newPos));
            gl_FragColor = vec4(oldArea / newArea * 0.2, 1.0, 0.0, 0.0);

            vec3 refractedLight = refract(-light, vec3(0.0, 1.0, 0.0), IOR_AIR / IOR_WATER);            
            /* compute a blob shadow and make sure we only draw a shadow if the player is blocking the light */
            vec3 dir = (sphereCenter - newPos) / sphereRadius;
            vec3 area = cross(dir, refractedLight);
            float shadow = dot(area, area);
            float dist = dot(dir, -refractedLight);
            shadow = 1.0 + (shadow - 1.0) / (0.05 + dist * 0.025);
            shadow = clamp(1.0 / (1.0 + exp(-shadow)), 0.0, 1.0);
            shadow = mix(1.0, shadow, clamp(dist * 2.0, 0.0, 1.0));
            gl_FragColor.g = shadow;
            
            /* shadow for the rim of the pool */
            vec2 t = intersectCube(newPos, -refractedLight, vec3(-1.0, -poolHeight, -1.0), vec3(1.0, 2.0, 1.0));
            gl_FragColor.r *= 1.0 / (1.0 + exp(-200.0 / (1.0 + 10.0 * (t.y - t.x)) * (newPos.y - refractedLight.y * t.y - 2.0 / 12.0)));
        }""", ctx)
        self.sky = Cubemap(local_textures("xpos.jpg"), local_textures("xneg.jpg"), local_textures("ypos.jpg"),
                           local_textures("ypos.jpg"), local_textures("zpos.jpg"), local_textures("zneg.jpg"), ctx)

    def update_caustics(self, matrices: Matrices, water: Water):
        self.caustic_texture.draw_to(self.ctx)
        self.ctx.clear()
        water.texture_a.use(0)
        self.caustics_shader.draw_mesh(self.water_mesh, matrices, unifroms={
            "light": self.light_dir,
            "water": 0,
            "sphereCenter": self.sphere_center,
            "sphereRadius": self.sphere_radius,
        })

    def render_water(self, matrices: Matrices, water: Water):
        water.texture_a.use(0)
        self.tile_texture.use(1)
        self.sky.use(2)
        self.caustic_texture.use(3)

        self.ctx.enable(moderngl.CULL_FACE)
        cull_face = ("front", "back")
        for idx in range(2):
            self.ctx.cull_face = cull_face[idx]
            self.water_shaders[idx].draw_mesh(self.water_mesh, matrices, unifroms={
                "light": self.light_dir,
                "water": 0,
                "tiles": 1,
                "sky": 2,
                "causticTex": 3,
                "eye": matrices.eye,
                "sphereCenter": self.sphere_center,
                "sphereRadius": self.sphere_radius,
            })
        self.ctx.disable(moderngl.CULL_FACE)

    def render_sphere(self, matrices: Matrices, water: Water):
        water.texture_a.use(0)
        self.caustic_texture.use(1)
        self.sphere_shader.draw_mesh(self.sphere_mesh, matrices, unifroms={
            "light": self.light_dir,
            "water": 0,
            "causticTex": 1,
            "sphereCenter": self.sphere_center,
            "sphereRadius": self.sphere_radius
        })

    def render_cude(self, matrices: Matrices, water: Water):
        self.ctx.enable(moderngl.CULL_FACE)
        water.texture_a.use(0)
        self.tile_texture.use(1)
        self.caustic_texture.use(2)
        self.cude_shader.draw_mesh(self.cube_mesh, matrices, unifroms={
            "light": self.light_dir,
            "water": 0,
            "tiles": 1,
            "causticTex": 2,
            "sphereCenter": self.sphere_center,
            "sphereRadius": self.sphere_radius
        })
        self.ctx.disable(moderngl.CULL_FACE)



"""
Global fixtures for moderngl tests.

We rely on Context.gc_mode = "auto" to clean up resources.
Mainly a global context is used it test, but some test needs a new context

* ctx_static: A global context that is reused for all tests
* ctx: Also a global context, but it is cleaned before and after each test
* ctx_new: A new context for each test

Context creation can be refined in _create_context if issues arise
"""
import pytest
import numpy as np
import moderngl
from glcontext import egl

VERSION_CODES = 430, 410, 330
_ctx = None
_fbo = None  # Fake framebuffer to avoid errors during transforms

@pytest.fixture(scope="session")
def ctx_static():
    """Session context"""
    return _get_context()


@pytest.fixture(scope="function")
def ctx():
    """
    Per function context.

    The same context is reused, but the context is cleaned before and after each test.
    """
    ctx = _get_context()
    error = ctx.error
    if error != "GL_NO_ERROR":
        raise RuntimeError("Context has error before use: {}".format(error))
    _clean_ctx(ctx)
    yield ctx
    error = ctx.error
    if error != "GL_NO_ERROR":
        raise RuntimeError("Context has error after use: {}".format(error))
    _clean_ctx(ctx)


@pytest.fixture(scope="function")
def ctx_new():
    """Returns a new context for each test"""
    ctx = _create_context()
    yield ctx
    ctx.release()


def _get_context():
    """Get the global test context"""
    global _ctx, _fbo
    if _ctx is None:
        _ctx = _create_context()
        _ctx.gc_mode = "auto"
        _fbo = _ctx.simple_framebuffer((2, 2))

    _ctx.__enter__()
    _fbo.use()
    return _ctx

def _create_context():
    """
    Create a new context.

    This is the only place context creation should happen.
    For now we just brute force context creation.
"""
    return moderngl.create_context(standalone=True, context=egl.create_context(glversion=330, mode="standalone"))

def _clean_ctx(ctx):
    """Clean the context"""
    # Reset the context
    ctx.blend_func = moderngl.DEFAULT_BLENDING
    ctx.blend_equation = moderngl.FUNC_ADD
    ctx.enable_only(moderngl.NOTHING)
    ctx.point_size = 1.0
    ctx.line_width = 1.0
    ctx.front_face = 'ccw'
    ctx.cull_face = 'back'
    ctx.wireframe = False
    ctx.provoking_vertex = moderngl.FIRST_VERTEX_CONVENTION
    ctx.polygon_offset = 0.0, 0.0    
    ctx.gc()


@pytest.fixture(scope="session")
def color_prog(ctx_static):
    """A simple program that renders a solid color."""
    return ctx_static.program(
        vertex_shader='''
            #version 330

            in vec2 in_vert;

            void main() {
                gl_Position = vec4(in_vert, 0.0, 1.0);
            }
        ''',
        fragment_shader='''
            #version 330

            out vec4 fragColor;
            uniform vec4 color;

            void main() {
                fragColor = color;
            }
        ''',
    )


@pytest.fixture(scope="session")
def ndc_quad(ctx_static):
    """Creates a buffer with an NDC quad."""
    quad = [
        -1.0,  1.0,
        -1.0, -1.0,
        1.0, 1.0,
        1.0, -1.0,
    ]
    return ctx_static.buffer(np.array(quad, dtype='f4'))



import random
import numpy as np
import moderngl
import moderngl_window as mglw
from light_gl import *
from water import *
from renderer import *


class WaterSimulation(mglw.WindowConfig):
    title = "WaterSimulation"
    gl_version = (3, 3)
    window_size = (1280, 720)
    aspect_ratio = 16 / 9
    resizable = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.matrix = Matrices()
        self.shader = Shader("""
                uniform sampler2D Heightmap;
                void main() {
                    vec2 coord = gl_Vertex.xy * 0.5 + 0.5;
                    float height = texture(Heightmap, coord).r * 0.5;
                    gl_Position = gl_ModelViewProjectionMatrix * vec4(gl_Vertex.xy, height, 1.0);
                }
            """, """
                uniform vec3 w_volor;
                void main() {
                    gl_FragColor = vec4(w_volor, 1.0);
                }
            """)
        self.shader.program["w_volor"] = glm.vec3(0.3, 0.5, 0.7)
        self.panel = MeshBuilder.panel(detail=64).build(self.ctx)

        self.water = Water()
        self._heavy = 5

    @property
    def heavy(self):
        return self._heavy

    @heavy.setter
    def heavy(self, value):
        self._heavy = min(10, max(0, value))

    def key_event(self, key, action, modifiers):
        if action == 1:
            if key == 265: # up arrow
                self.heavy += 1
            elif key == 264: # down arrow
                self.heavy -= 1

    def render(self, time, frame_time):
        angle = time * 0.2
        self.matrix.projection = glm.perspective(
            45.0, self.aspect_ratio, 0.1, 1000.0)
        self.matrix.view = glm.lookAt((np.cos(angle), np.sin(angle), 2.2),
                                      (0.0, 0.0, 0.0),
                                      (0.0, 0.0, 1.0),)

        if random.randint(0, 50) < self._heavy:
            strength = random.choice([1, -1]) * 0.4
            self.water.add_drop(random.random() * 2.0 - 1.0,
                                random.random() * 2.0 - 1.0, 0.04, strength)

        self.water.step_simulation()
        self.water.step_simulation()
        self.water.update_normals()

        self.ctx.enable(moderngl.DEPTH_TEST)
        self.ctx.screen.use()
        self.water.texture_a.use()
        self.shader.draw_mesh(self.panel, self.matrix, mode=moderngl.LINES)
        self.ctx.disable(moderngl.DEPTH_TEST)


if __name__ == '__main__':
    mglw.run_window_config(WaterSimulation, args=["--window", "glfw"])



# Example of shadow rendering (shadow mapping)
# Description: This example shows how you can draw a shadow
#              from a single light source.
# Note: The left-hand rule coordinate system is used.

import moderngl as gl
import moderngl_window as glw
import struct
import math


# Create window
window_cls = glw.get_local_window_cls('pyglet')
window = window_cls(
    size=(1024, 512), fullscreen=False,
    title='Shadow rendering (shadow mapping) example',
    resizable=False, vsync=True, gl_version=(3, 3)
)
ctx = window.ctx
glw.activate_context(window, ctx=ctx)
window.clear()
window.swap_buffers()

# Window program
window_program = ctx.program(
    vertex_shader='''
    #version 330

    in vec3 in_vert;
    in vec4 in_color;

    uniform float camera_aspect_ratio;
    uniform vec3 camera_position;
    uniform mat3 camera_rotation_matrix;
    uniform mat4 camera_projection_matrix;

    out vec3 o_vert;
    out vec4 o_color;

    void main() {
        vec3 new_position = camera_rotation_matrix*(in_vert-camera_position);
        gl_Position = camera_projection_matrix*vec4(
            new_position.x, new_position.y*camera_aspect_ratio,
            new_position.z, 1.0
        );

        o_vert = new_position;
        o_color = in_color;
    }
    ''',
    fragment_shader='''
    #version 330

    in vec3 o_vert;
    in vec4 o_color;

    uniform float light_aspect_ratio;
    uniform vec2 light_clip_space;
    uniform float light_perspective;
    uniform vec3 light_position;
    uniform mat3 light_rotation_matrix;

    uniform vec3 camera_position;
    uniform mat3 camera_revers_rotation_matrix;

    uniform float shadow_bias;

    uniform sampler2D light_depthI;

    out vec4 f_color;

    void main() {
        f_color = o_color;

        // Camera pixels position //
        vec3 pixel_coords = o_vert;
        ////////////

        // ROTATION 3D PROG moving prog //
        // Convert to light pixel coords //
        pixel_coords = camera_revers_rotation_matrix*
              pixel_coords;
        pixel_coords += camera_position;
        vec3 pixel_world_coords = pixel_coords;  // Pixel world coordinates
        pixel_coords -= light_position;
        pixel_coords = light_rotation_matrix*
              pixel_coords;
        /////////////////

        // Convert .xy to 2D coords //
        float distance = length(pixel_coords);
        // ^^ From light to pixel 3D distance ^^

        pixel_coords.y*=light_aspect_ratio;
        pixel_coords.xy = vec2(pixel_coords.xy/
                (pixel_coords.z+(1.0-pixel_coords.z)*(1.0-light_perspective))+1.0)/2.0;
        ///////////


        // Get depth
        float depth = texture(light_depthI, pixel_coords.xy).x;
        depth = 2.0 * depth - 1.0;

        vec2 lcs = light_clip_space;

        depth = (
            (2.0*lcs.x*lcs.y/(lcs.y+lcs.x-depth*(lcs.y-lcs.x)))*
                light_perspective+
            ((depth+(lcs.y+lcs.x)/(lcs.y-lcs.x))*(lcs.y - lcs.x))/2*
                (1.0-light_perspective)
        );

        // Compare depth
        if (!(
             pixel_coords.x > 0.0 && pixel_coords.y > 0.0
            && pixel_coords.x < 1.0 && pixel_coords.y < 1.0
            && pixel_coords.z > 0.0
            && pixel_coords.z < depth+shadow_bias
        )) {
            f_color.xyz = f_color.xyz*0.3;
        }
    }
    '''
)

# Light program
light_program = ctx.program(
    vertex_shader='''
    #version 330

    in vec3 in_vert;
    in vec4 in_color;

    uniform float light_aspect_ratio;
    uniform vec3 light_position;
    uniform mat3 light_rotation_matrix;
    uniform mat4 light_projection_matrix;

    out vec4 o_color;

    void main() {
        vec3 new_position = light_rotation_matrix*(in_vert-light_position);
        gl_Position = light_projection_matrix*vec4(
            new_position.x, new_position.y*light_aspect_ratio,
            new_position.z, 1.0
        );

        o_color = in_color;
    }
    ''',
    fragment_shader='''
    #version 330

    in vec4 o_color;
    out vec4 f_color;

    void main() {
        f_color = o_color;
    }
    '''
)

# Preparing scene
vertex = [
    -.5, -.1, -.5,
    -.5, -.1, .5,
    .5, -.1, .5,

    .5, -.1, -.5,
    .5, -.1, .5,
    -.5, -.1, -.5,

    .2, -.1, -.2,
    .2, -.1, .2,
    -.2, .05, .2,
]
vertex_buf = ctx.buffer(struct.pack(f'{len(vertex)}f', *vertex))

colors = [
    .5, .5, .5, 1,
    .5, .5, .5, 1,
    .5, .5, .5, 1,

    .5, .5, .5, 1,
    .5, .5, .5, 1,
    .5, .5, .5, 1,

    1, 0, 0, 1,
    0, 1, 0, 1,
    0, 0, 1, 1,
]
colors_buf = ctx.buffer(struct.pack(f'{len(colors)}f', *colors))

# Prepair render
window_vao = ctx.vertex_array(window_program, [
    (vertex_buf, '3f', 'in_vert'),
    (colors_buf, '4f', 'in_color')
])
window_fbo = ctx.framebuffer(
    color_attachments=[ctx.texture(window.size, 4)],
    depth_attachment=ctx.depth_texture(window.size)
)
window_fbo.color_attachments[0].repeat_x = False
window_fbo.color_attachments[0].repeat_y = False
window_fbo.depth_attachment.repeat_x = False
window_fbo.depth_attachment.repeat_y = False
window_fbo.depth_attachment.compare_func = ''


light_vao = ctx.vertex_array(light_program, [
    (vertex_buf, '3f', 'in_vert'),
    (colors_buf, '4f', 'in_color')
])
light_fbo = ctx.framebuffer(
    color_attachments=[ctx.texture(window.size, 4)],
    depth_attachment=ctx.depth_texture(window.size)
)
light_fbo.color_attachments[0].repeat_x = False
light_fbo.color_attachments[0].repeat_y = False
light_fbo.depth_attachment.repeat_x = False
light_fbo.depth_attachment.repeat_y = False
light_fbo.depth_attachment.filter = gl.LINEAR, gl.LINEAR_MIPMAP_LINEAR
light_fbo.depth_attachment.compare_func = ''


# Define matrix methods
def get_projection_matrix(near, far, perspective=1):
    perspective_projection = (
        1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, (far+near)/(far-near), 1,
        0, 0, -(2*far*near)/(far-near), 0
    )
    orthographic_projection = (
        1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 2/(far-near), 0,
        0, 0, -(far+near)/(far-near), 1
    )

    result_projection = []
    for x, y in zip(perspective_projection, orthographic_projection):
        result_projection.append(x*perspective+y*(1-perspective))

    return result_projection


def get_rotation_matrix(rx, ry, rz, revers=True):
    rx *= (math.pi / 180)
    ry *= (math.pi / 180)
    rz *= (math.pi / 180)
    if not revers:
        srx, crx = math.sin(rx), math.cos(rx)
        sry, cry = math.sin(-ry), math.cos(-ry)
        srz, crz = math.sin(-rz), math.cos(-rz)
        rotation_matrix = (
            cry * crz - sry * srx * srz,
            cry * srz + sry * srx * crz,
            -sry * crx,

            -crx * srz,
            crx * crz,
            srx,

            sry * crz + cry * srx * srz,
            sry * srz - cry * srx * crz,
            cry * crx
        )
    else:
        srx, crx = math.sin(-rx), math.cos(-rx)
        sry, cry = math.sin(ry), math.cos(ry)
        srz, crz = math.sin(rz), math.cos(rz)
        rotation_matrix = (
            crz * cry + srz * srx * sry,
            srz * crx,
            srz * srx * cry - crz * sry,

            crz * srx * sry - srz * cry,
            crz * crx,
            srz * sry + crz * srx * cry,

            crx * sry,
            -srx,
            crx * cry
        )

    return rotation_matrix


#####################################################################
# Define constants
far = 10  # Camera projection far
near = 0.001  # Camera projection near
perpective = 1  # Select perpective/orthographic projection
position = [0., 0.1, -0.4]  # Camera position
rotate = [-20, 0, 0]  # Camera angle

light_far = 10  # Light projection far
light_near = 0.001  # Light projection near
light_perspective = 0  # Select perpective/orthographic projection
light_position = [0, 0.4, 0]  # Light position
light_rotate = [-90, 0, 0]  # Light angle

shadow_bias = 0.001  # Allowed shadow offset

view_as_light = False  # Show what the light sees
#####################################################################

# Filling uniforms
window_vao.program['camera_aspect_ratio'] = \
    window_fbo.size[0]/window_fbo.size[1]
window_vao.program['camera_position'] = position
window_vao.program['camera_projection_matrix'] = \
    get_projection_matrix(near, far, perpective)
window_vao.program['camera_rotation_matrix'] = \
    get_rotation_matrix(*rotate, revers=False)
window_vao.program['light_aspect_ratio'] = light_fbo.size[0]/light_fbo.size[1]
window_vao.program['light_clip_space'] = light_near, light_far
window_vao.program['light_perspective'] = light_perspective
window_vao.program['light_position'] = light_position
window_vao.program['light_rotation_matrix'] = \
    get_rotation_matrix(*light_rotate, revers=False)
window_vao.program['camera_revers_rotation_matrix'] = \
    get_rotation_matrix(*rotate, revers=True)

light_vao.program['light_aspect_ratio'] = light_fbo.size[0]/light_fbo.size[1]
light_vao.program['light_position'] = light_position
light_vao.program['light_projection_matrix'] = \
    get_projection_matrix(light_near, light_far, light_perspective)
light_vao.program['light_rotation_matrix'] = \
    get_rotation_matrix(*light_rotate, revers=False)

window_vao.program['shadow_bias'] = shadow_bias


# Render loop
ctx.enable(gl.DEPTH_TEST)
ctx.depth_func = '<'

while not window.is_closing:
    light_fbo.use()
    light_fbo.clear()
    light_vao.render(vertices=vertex_buf.size // 4 // 3)

    window_fbo.use()
    window_fbo.clear()
    light_fbo.depth_attachment.use(location=0)
    window_vao.render(vertices=vertex_buf.size // 4 // 3)

    ctx.copy_framebuffer(
        window.fbo,
        light_fbo if view_as_light else window_fbo
    )

    window.swap_buffers()



# WebGL Water
# https://madebyevan.com/webgl-water/
# Copyright 2011 Evan Wallace
# Released under the MIT license

from typing import Self
import glm
import math
from light_gl import Matrices


class HitTest:
    def __init__(self, t: float, hit: glm.vec3, normal: glm.vec3) -> None:
        self.t = t
        self.hit = hit
        self.normal = normal

    def merge_with(self, other: Self):
        if other.t > 0 and other.t < self.t:
            self.t = other.t
            self.hit = other.hit
            self.normal = other.normal

    def __repr__(self) -> str:
        return f"t: {self.t} hit: {self.hit} norm: {self.normal}"


class RayTracer:
    def __init__(self, viewport: tuple[float], matrices: Matrices) -> None:
        model_view_matrix = matrices.model_view
        self.eye = matrices.eye
        self.viewport = viewport

        min_x = viewport[0]
        max_x = min_x + viewport[2]
        min_y = viewport[1]
        max_y = min_y + viewport[3]

        self.ray00 = glm.unProject(
            (min_x, min_y, 1), model_view_matrix, matrices.projection, viewport) - self.eye
        self.ray10 = glm.unProject(
            (max_x, min_y, 1), model_view_matrix, matrices.projection, viewport) - self.eye
        self.ray01 = glm.unProject(
            (min_x, max_y, 1), model_view_matrix, matrices.projection, viewport) - self.eye
        self.ray11 = glm.unProject(
            (max_x, max_y, 1), model_view_matrix, matrices.projection, viewport) - self.eye

    def get_ray_for_pixel(self, x, y) -> glm.vec3:
        x = (x - self.viewport[0]) / self.viewport[2]
        y = 1.0 - (y - self.viewport[1]) / self.viewport[3]
        ray0 = glm.lerp(self.ray00, self.ray10, x)
        ray1 = glm.lerp(self.ray01, self.ray11, x)
        return glm.normalize(glm.lerp(ray0, ray1, y))

    def hit_test_sphere(origin: glm.vec3, ray: glm.vec3, center: glm.vec3, radius: float) -> HitTest:
        offset = origin - center
        a = glm.dot(ray, ray)
        b = 2 * glm.dot(ray, offset)
        c = glm.dot(offset, offset) - radius * radius
        discriminant = b * b - 4 * a * c
        if discriminant > 0:
            t = (-b - math.sqrt(discriminant)) / (2*a)
            hit = origin + ray * t
            return HitTest(t, hit, (hit - center) / radius)
        return None

    def hit_test_triangle(origin: glm.vec3, ray: glm.vec3, vert: tuple[glm.vec3]):
        ab = vert[1] - vert[0]
        ac = vert[2] - vert[0]
        normal = glm.normalize(glm.cross(ab, ac))
        t = glm.dot(normal, vert[0] - origin) / glm.dot(normal, ray)
        if t > 0:
            hit = origin + (ray * t)
            to_hit = hit - vert[0]
            # uv = glm.inverse((glm.mat3(ac, ab, normal))) * to_hit
            dot00 = glm.dot(ac, ac)
            dot01 = glm.dot(ab, ac)
            dot11 = glm.dot(ab, ab)
            dot02 = glm.dot(ac, to_hit)
            dot12 = glm.dot(ab, to_hit)
            divide = dot00 * dot11 - dot01 * dot01
            u = (dot11 * dot02 - dot01 * dot12) / divide
            v = (dot00 * dot12 - dot01 * dot02) / divide
            print("uv", u, v)
            if u >= 0 and v >= 0 and (u + v) <= 1:
                return HitTest(t, hit, normal)
        return None



import moderngl
import numpy as np

ctx = moderngl.create_context(standalone=True)

prog = ctx.program(
    vertex_shader="""
        #version 330

        in vec2 in_vert;
        in vec3 in_color;

        out vec3 v_color;

        void main() {
            v_color = in_color;
            gl_Position = vec4(in_vert, 0.0, 1.0);
        }
    """,
    fragment_shader="""
        #version 330

        in vec3 v_color;

        out vec3 f_color;

        void main() {
            f_color = v_color;
        }
    """,
)

x = np.linspace(-1.0, 1.0, 50)
y = np.random.rand(50) - 0.5
r = np.zeros(50)
g = np.ones(50)
b = np.zeros(50)

vertices = np.dstack([x, y, r, g, b])

vbo = ctx.buffer(vertices.astype("f4").tobytes())
vao = ctx.vertex_array(prog, vbo, "in_vert", "in_color")



import array
import time

import layered_window
import moderngl

mem = layered_window.init((400, 400), title='Animation')
# mem = layered_window.init((400, 400), title='Animation', tool_window=True, always_on_top=True)

ctx = moderngl.create_context(standalone=True)

image = ctx.texture((400, 400), components=4)
framebuffer = ctx.framebuffer(color_attachments=[image])

triangle = array.array('f', [
    1.0, 0.0, 1.0, 0.0, 0.0, 0.5,
    -0.5, 0.86, 0.0, 1.0, 0.0, 0.5,
    -0.5, -0.86, 0.0, 0.0, 1.0, 0.5,
])

vertex_buffer = ctx.buffer(triangle)

program = ctx.program(
    vertex_shader='''
        #version 330 core

        uniform vec2 scale;
        uniform float rotation;

        layout (location = 0) in vec2 in_vertex;
        layout (location = 1) in vec4 in_color;

        out vec4 v_color;

        void main() {
            float r = rotation * (0.5 + float(gl_InstanceID) * 0.05);
            mat2 rot = mat2(cos(r), sin(r), -sin(r), cos(r));
            gl_Position = vec4((rot * in_vertex) * scale, 0.0, 1.0);
            v_color = in_color;
        }
    ''',
    fragment_shader='''
        #version 330 core

        in vec4 v_color;

        layout (location = 0) out vec4 out_color;

        void main() {
            out_color = vec4(v_color);
            out_color.rgb = pow(out_color.rgb, vec3(1.0 / 2.2));
        }
    ''',
)

ctx.enable_only(moderngl.BLEND)
ctx.blend_func = (ctx.SRC_ALPHA, ctx.ONE_MINUS_SRC_ALPHA, ctx.SRC_ALPHA, ctx.ONE)
vao = ctx.vertex_array(program, [
    (vertex_buffer, '2f 4f', 'in_vertex', 'in_color'),
])
vao.vertices = 3
vao.instances = 10

start_time = time.perf_counter()

while True:
    now = time.perf_counter() - start_time
    framebuffer.use()
    framebuffer.clear(0.0, 0.0, 0.0, 0.0)
    program['scale'].value = (0.8, 0.8)
    program['rotation'].value = now
    vao.render()
    framebuffer.read_into(mem, components=4)
    layered_window.update()



import math
import os
import sys

import glm
import moderngl
import pygame
from objloader import Obj
from PIL import Image

os.environ['SDL_WINDOWS_DPI_AWARENESS'] = 'permonitorv2'

pygame.init()
pygame.display.set_mode((800, 800), flags=pygame.OPENGL | pygame.DOUBLEBUF, vsync=True)


class ColorMaterial:
    def __init__(self):
        self.ctx = moderngl.get_context()
        self.program = self.ctx.program(
            vertex_shader='''
                #version 330 core

                uniform mat4 camera;
                uniform vec3 position;
                uniform float scale;

                layout (location = 0) in vec3 in_vertex;
                layout (location = 1) in vec3 in_normal;
                layout (location = 2) in vec2 in_uv;

                out vec3 v_vertex;
                out vec3 v_normal;
                out vec2 v_uv;

                void main() {
                    v_vertex = position + in_vertex * scale;
                    v_normal = in_normal;
                    v_uv = in_uv;

                    gl_Position = camera * vec4(v_vertex, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core

                uniform vec3 light_direction;
                uniform vec3 color;

                in vec3 v_vertex;
                in vec3 v_normal;
                in vec2 v_uv;

                layout (location = 0) out vec4 out_color;

                void main() {
                    out_color = vec4(color, 1.0);
                    float lum = dot(normalize(v_normal), normalize(light_direction));
                    out_color.rgb *= max(lum, 0.0) * 0.5 + 0.5;
                }
            ''',
        )

        self.color = (1.0, 1.0, 1.0)

    def use(self):
        self.program['color'] = self.color

    def vertex_array(self, buffer):
        return self.ctx.vertex_array(self.program, [(buffer, '3f 3f 8x', 'in_vertex', 'in_normal')])


class TextureMaterial:
    def __init__(self, texture):
        self.texture = texture
        self.ctx = moderngl.get_context()
        self.program = self.ctx.program(
            vertex_shader='''
                #version 330 core

                uniform mat4 camera;
                uniform vec3 position;
                uniform float scale;

                layout (location = 0) in vec3 in_vertex;
                layout (location = 1) in vec3 in_normal;
                layout (location = 2) in vec2 in_uv;

                out vec3 v_vertex;
                out vec3 v_normal;
                out vec2 v_uv;

                void main() {
                    v_vertex = position + in_vertex * scale;
                    v_normal = in_normal;
                    v_uv = in_uv;

                    gl_Position = camera * vec4(v_vertex, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core

                uniform sampler2D Texture;
                uniform vec3 light_direction;

                in vec3 v_vertex;
                in vec3 v_normal;
                in vec2 v_uv;

                layout (location = 0) out vec4 out_color;

                void main() {
                    out_color = texture(Texture, v_uv);
                    float lum = dot(normalize(v_normal), normalize(light_direction));
                    out_color.rgb *= max(lum, 0.0) * 0.5 + 0.5;
                }
            ''',
        )

    def use(self):
        self.texture.use()

    def vertex_array(self, buffer):
        return self.ctx.vertex_array(self.program, [(buffer, '3f 3f 2f', 'in_vertex', 'in_normal', 'in_uv')])


class ImageTexture:
    def __init__(self, path):
        self.ctx = moderngl.get_context()

        img = Image.open(path).convert('RGBA')
        self.texture = self.ctx.texture(img.size, 4, img.tobytes())
        self.sampler = self.ctx.sampler(texture=self.texture)

    def use(self):
        self.sampler.use()


class ModelGeometry:
    def __init__(self, path):
        self.ctx = moderngl.get_context()

        obj = Obj.open(path)
        self.vbo = self.ctx.buffer(obj.pack('vx vy vz nx ny nz tx ty'))


class Mesh:
    def __init__(self, material, geometry):
        self.ctx = moderngl.get_context()
        self.vao = material.vertex_array(geometry.vbo)
        self.material = material

    def render(self, position, scale):
        self.material.use()
        self.vao.program['position'] = position
        self.vao.program['scale'] = scale
        self.vao.render()


class Scene:
    def __init__(self):
        self.ctx = moderngl.get_context()

        self.texture = ImageTexture('examples/data/textures/crate.png')

        self.color_material = ColorMaterial()
        self.texture_material = TextureMaterial(self.texture)

        self.car_geometry = ModelGeometry('examples/data/models/lowpoly_toy_car.obj')
        self.car = Mesh(self.color_material, self.car_geometry)

        self.crate_geometry = ModelGeometry('examples/data/models/crate.obj')
        self.crate = Mesh(self.texture_material, self.crate_geometry)

    def camera_matrix(self):
        now = pygame.time.get_ticks() / 1000.0
        eye = (math.cos(now), math.sin(now), 0.5)
        proj = glm.perspective(45.0, 1.0, 0.1, 1000.0)
        look = glm.lookAt(eye, (0.0, 0.0, 0.0), (0.0, 0.0, 1.0))
        return proj * look

    def render(self):
        camera = self.camera_matrix()

        self.ctx.clear()
        self.ctx.enable(self.ctx.DEPTH_TEST)

        self.color_material.program['camera'].write(camera)
        self.color_material.program['light_direction'] = (1.0, 2.0, 3.0)

        self.texture_material.program['camera'].write(camera)
        self.texture_material.program['light_direction'] = (1.0, 2.0, 3.0)

        self.crate.render((0.0, 0.0, 0.0), 0.2)

        self.color_material.color = (1.0, 0.0, 0.0)
        self.car.render((-0.4, 0.0, 0.0), 0.2)

        self.color_material.color = (0.0, 0.0, 1.0)
        self.car.render((0.4, 0.0, 0.0), 0.2)


scene = Scene()

while True:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            pygame.quit()
            sys.exit()

    scene.render()

    pygame.display.flip()



# WebGL Water
# https://madebyevan.com/webgl-water/
# Copyright 2011 Evan Wallace
# Released under the MIT license

import moderngl
from light_gl import *


class Water:
    _vert_shader = """
    varying vec2 coord;
    void main() {
        coord = gl_Vertex.xy * 0.5 + 0.5;
        gl_Position = vec4(gl_Vertex.xyz, 1.0);
    }
    """

    _drop_frag_shader = """
    const float PI = 3.141592653589793;
    uniform sampler2D texture;
    uniform vec2 center;
    uniform float radius;
    uniform float strength;
    varying vec2 coord;
    void main() {
        /* get vertex info */
        vec4 info = texture2D(texture, coord);
        /* add the drop to the height */
        float drop = max(0.0, 1.0 - length(center * 0.5 + 0.5 - coord) / radius);
        drop = 0.5 - cos(drop * PI) * 0.5;
        info.r += drop * strength;
        gl_FragColor = info;
    }
    """

    _update_frag_shader = """
    uniform sampler2D texture;
    uniform vec2 delta;
    varying vec2 coord;

    vec4 water_sample(vec2 coord, vec4 info) {
        // vec2 center = vec2(0.5, 0.4);
        // float radius = 0.10;  
        // if (length(coord.xy - center) < radius) {
        //     return info;
        // }
        return texture2D(texture, coord);
    }

    void main() {
        /* get vertex info */
        vec4 info = water_sample(coord, vec4(0.5));
        
        /* calculate average neighbor height */
        vec2 dx = vec2(delta.x, 0.0);
        vec2 dy = vec2(0.0, delta.y);
        float average = (
            water_sample(coord - dx, info).r +
            water_sample(coord - dy, info).r +
            water_sample(coord + dx, info).r +
            water_sample(coord + dy, info).r
        ) * 0.25;
        
        /* change the velocity to move toward the average */
        info.g += (average - info.r) * 2.0;
        
        /* attenuate the velocity a little so waves do not last forever */
        info.g *= 0.995;
        
        /* move the vertex along the velocity */
        info.r += info.g;

        gl_FragColor = vec4(info.rg, 0.0, 0.0);
    }
    """

    _normal_frag_shader = """
    uniform sampler2D texture;
    uniform vec2 delta;
    varying vec2 coord;
    void main() {
        /* get vertex info */
        vec4 info = texture2D(texture, coord);
        
        /* update the normal */
        vec3 dx = vec3(delta.x, texture2D(texture, vec2(coord.x + delta.x, coord.y)).r - info.r, 0.0);
        vec3 dy = vec3(0.0, texture2D(texture, vec2(coord.x, coord.y + delta.y)).r - info.r, delta.y);
        info.ba = normalize(cross(dy, dx)).xz;
        
        gl_FragColor = info;
    }
    """

    _sphere_frag_shader = """
    uniform sampler2D texture;
    uniform vec3 oldCenter;
    uniform vec3 newCenter;
    uniform float radius;
    varying vec2 coord;
    
    float volumeInSphere(vec3 center) {
      vec3 toCenter = vec3(coord.x * 2.0 - 1.0, 0.0, coord.y * 2.0 - 1.0) - center;
      float t = length(toCenter) / radius;
      float dy = exp(-pow(t * 1.5, 6.0));
      float ymin = min(0.0, center.y - dy);
      float ymax = min(max(0.0, center.y + dy), ymin + 2.0 * dy);
      return (ymax - ymin) * 0.1;
    }
    
    void main() {
      /* get vertex info */
      vec4 info = texture2D(texture, coord);
      
      /* add the old volume */
      info.r += volumeInSphere(oldCenter);
      
      /* subtract the new volume */
      info.r -= volumeInSphere(newCenter);
      
      gl_FragColor = info;
    }
    """

    def __init__(self, texture_size=(256, 256), ctx: moderngl.Context = None) -> None:
        self.ctx = ctx or None
        self.panel = MeshBuilder.panel().build(self.ctx)
        self.texture_a = RawTexture(
            texture_size, pixel_fmt="RGBA", ctx=self.ctx, dtype="f4")
        self.texture_b = RawTexture(
            texture_size, pixel_fmt="RGBA", ctx=self.ctx, dtype="f4")
        self.delta = glm.vec2(
            1.0 / texture_size[0], 1.0 / texture_size[1])
        self.drop_shader = Shader(
            Water._vert_shader, Water._drop_frag_shader, self.ctx)
        self.update_shader = Shader(
            Water._vert_shader, Water._update_frag_shader, self.ctx)
        self.normal_shader = Shader(
            Water._vert_shader, Water._normal_frag_shader, self.ctx)
        self.sphere_shader = Shader(
            Water._vert_shader, Water._sphere_frag_shader, self.ctx)

    def swap(self):
        temp_texture = self.texture_b
        self.texture_b = self.texture_a
        self.texture_a = temp_texture

    def add_drop(self, x: float, y: float, radius: float, strength: float):
        self.texture_b.draw_to(self.ctx)
        self.texture_a.use()
        self.drop_shader.draw_mesh(self.panel, unifroms={
            "center": glm.vec2(x, y),
            "radius": radius,
            "strength": strength
        })
        self.swap()

    def move_sphere(self, old_center, new_center, radius):
        self.texture_b.draw_to(self.ctx)
        self.texture_a.use()
        self.sphere_shader.draw_mesh(self.panel, unifroms={
            "oldCenter": old_center,
            "newCenter": new_center,
            "radius": radius
        })
        self.swap()

    def step_simulation(self):
        self.texture_b.draw_to(self.ctx)
        self.texture_a.use()
        self.update_shader.draw_mesh(self.panel, unifroms={
            "delta": self.delta
        })
        self.swap()
    
    def update_normals(self):
        self.texture_b.draw_to(self.ctx)
        self.texture_a.use()
        self.normal_shader.draw_mesh(self.panel, unifroms={
            "delta": self.delta
        })
        self.swap()



import moderngl_window as glw
import numpy as np

window_cls = glw.get_local_window_cls('pyglet')
window = window_cls(
    size=(512, 512), fullscreen=False, title='ModernGL Window',
    resizable=False, vsync=True, gl_version=(3, 3)
)
ctx = window.ctx
glw.activate_context(window, ctx=ctx)
window.clear()
window.swap_buffers()

prog = ctx.program(
    vertex_shader="""
        #version 330

        in vec2 in_vert;
        in vec3 in_color;

        out vec3 v_color;

        void main() {
            v_color = in_color;
            gl_Position = vec4(in_vert, 0.0, 1.0);
        }
    """,
    fragment_shader="""
        #version 330

        in vec3 v_color;

        out vec3 f_color;

        void main() {
            f_color = v_color;
        }
    """,
)

x = np.linspace(-1.0, 1.0, 50)
y = np.random.rand(50) - 0.5
r = np.zeros(50)
g = np.ones(50)
b = np.zeros(50)

vertices = np.dstack([x, y, r, g, b])

vbo = ctx.buffer(vertices.astype("f4").tobytes())
vao = ctx.vertex_array(prog, vbo, "in_vert", "in_color")

fbo = ctx.framebuffer(
    color_attachments=[ctx.texture((512, 512), 3)]
)


while not window.is_closing:
    fbo.use()
    fbo.clear(0.0, 0.0, 0.0, 1.0)
    vao.render(gl.LINE_STRIP)

    ctx.copy_framebuffer(window.fbo, fbo)

    window.swap_buffers()



'''
    Renders a color triangle with imageStore
'''

from _example import Example


class HelloWorld(Example):
    gl_version = (4, 5)
    title = "Empty Framebuffer"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 450 core
                out vec3 v_color;
                vec2 positions[3] = vec2[](
                    vec2(0.0, 0.8),
                    vec2(-0.6, -0.8),
                    vec2(0.6, -0.8)
                );
                vec3 colors[3] = vec3[](
                    vec3(1.0, 0.0, 0.0),
                    vec3(0.0, 1.0, 0.0),
                    vec3(0.0, 0.0, 1.0)
                );
                void main() {
                    gl_Position = vec4(positions[gl_VertexID], 0.0, 1.0);
                    v_color = colors[gl_VertexID];
                }
            ''',
            fragment_shader='''
                #version 450 core
                in vec3 v_color;
                layout(rgba8, binding = 0) writeonly uniform image2D output_image;
                void main() {
                    imageStore(output_image, ivec2(gl_FragCoord.xy), vec4(v_color, 1.0));
                }
            ''',
        )

        self.prog['output_image'] = 0

        self.texture = self.ctx.texture(self.window_size, components=4)
        self.texture.bind_to_image(0, read=False, write=True)

        self.empty_framebuffer = self.ctx.empty_framebuffer(self.window_size)
        self.vao = self.ctx.vertex_array(self.prog, [])
        self.vao.scope = self.ctx.scope(self.empty_framebuffer)
        self.vao.vertices = 3

        self.fbo = self.ctx.framebuffer([self.texture])

    def render(self, time, frame_time):
        self.fbo.clear(1.0, 1.0, 1.0, 1.0)
        self.vao.render()
        self.ctx.memory_barrier()
        self.ctx.copy_framebuffer(self.ctx.screen, self.fbo)


if __name__ == '__main__':
    HelloWorld.run()



import math
import os
import sys

import glm
import moderngl
import pygame
from objloader import Obj
from PIL import Image

os.environ['SDL_WINDOWS_DPI_AWARENESS'] = 'permonitorv2'

pygame.init()
pygame.display.set_mode((800, 800), flags=pygame.OPENGL | pygame.DOUBLEBUF, vsync=True)


class Scene:
    def __init__(self):
        self.ctx = moderngl.get_context()

        self.program = self.ctx.program(
            vertex_shader='''
                #version 330 core

                uniform mat4 camera;

                layout (location = 0) in vec3 in_vertex;
                layout (location = 1) in vec3 in_normal;
                layout (location = 2) in vec2 in_uv;

                out vec3 v_vertex;
                out vec3 v_normal;
                out vec2 v_uv;

                void main() {
                    v_vertex = in_vertex;
                    v_normal = in_normal;
                    v_uv = in_uv;

                    gl_Position = camera * vec4(v_vertex, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core

                uniform sampler2D Texture;
                uniform vec3 light_direction;

                in vec3 v_vertex;
                in vec3 v_normal;
                in vec2 v_uv;

                layout (location = 0) out vec4 out_color;

                void main() {
                    float lum = max(dot(normalize(v_normal), normalize(light_direction)), 0.0) * 0.5 + 0.5;
                    out_color = texture(Texture, v_uv) * lum;
                }
            ''',
        )

        self.texture = self.load_texture('examples/data/textures/crate.png')
        self.vertex_buffer = self.load_model('examples/data/models/crate.obj')

        self.sampler = self.ctx.sampler(texture=self.texture)
        self.vertex_array = self.ctx.vertex_array(self.program, [
            (self.vertex_buffer, '3f 3f 2f', 'in_vertex', 'in_normal', 'in_uv'),
        ])

    def load_model(self, path):
        obj = Obj.open(path)
        return self.ctx.buffer(obj.pack('vx vy vz nx ny nz tx ty'))

    def load_texture(self, path):
        from OpenGL import GL
        img = Image.open(path).convert('RGBA')
        width, height = img.size
        pixels = img.tobytes()
        texture = GL.glGenTextures(1)
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, texture)
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA8, width, height, 0, GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, pixels)
        return self.ctx.external_texture(texture, (width, height), 4, 0, 'f1')

    def camera_matrix(self):
        now = pygame.time.get_ticks() / 1000.0
        eye = (math.cos(now) * 3.0, math.sin(now) * 3.0, 1.75)
        proj = glm.perspective(45.0, 1.0, 0.1, 1000.0)
        look = glm.lookAt(eye, (0.0, 0.0, 0.5), (0.0, 0.0, 1.0))
        return proj * look

    def render(self):
        self.ctx.clear()
        self.ctx.enable_only(self.ctx.DEPTH_TEST)
        self.program['light_direction'] = (1.0, 2.0, 3.0)
        self.program['camera'].write(self.camera_matrix())
        self.sampler.use()
        self.vertex_array.render()


scene = Scene()

while True:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            pygame.quit()
            sys.exit()

    scene.render()

    pygame.display.flip()



import moderngl
import numpy as np

from PIL import Image

ctx = moderngl.create_context(standalone=True)

prog = ctx.program(
    vertex_shader="""
        #version 330

        in vec2 in_vert;
        in vec3 in_color;

        uniform float scale;

        out vec3 v_color;

        void main() {
            v_color = in_color;
            gl_Position = vec4(in_vert * scale, 0.0, 1.0);
        }
    """,
    fragment_shader="""
        #version 330

        in vec3 v_color;

        out vec3 f_color;

        void main() {
            f_color = v_color;
        }
    """,
)

vertices = np.asarray([

    -0.75, -0.75,  1, 0, 0,
    0.75, -0.75,  0, 1, 0,
    0.0, 0.649,  0, 0, 1

], dtype='f4')

vbo = ctx.buffer(vertices.tobytes())
vao = ctx.vertex_array(prog, vbo, "in_vert", "in_color")

fbo = ctx.framebuffer(
    color_attachments=[ctx.texture((512, 512), 3)]
)
fbo.use()
fbo.clear(0.0, 0.0, 0.0, 1.0)

vao.program['scale'] = 2

vao.render()  # "mode" is moderngl.TRIANGLES by default

Image.frombytes(
    "RGB", fbo.size, fbo.color_attachments[0].read(),
    "raw", "RGB", 0, -1
).show()



"""
Generates buffer data on the gpu using transform feedback.
We use an empty VertexArray running it vertices times outputting data to the buffer.

More fancy calculations can of course be done. This example just shows the concept.
"""
import struct
from _example import Example


class GenerateData(Example):
    title = "Generate Data"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                out vec3 pos;
                out vec3 color;

                void main() {
                    pos = vec3(gl_VertexID * 2, gl_VertexID * 2, gl_VertexID * 2);
                    color = vec3(fract(gl_VertexID * 12.344), fract(gl_VertexID * 34.111), fract(gl_VertexID * 7.244));
                }
            ''',
            varyings=['pos', 'color'],
        )

        N = 4
        self.buffer = self.ctx.buffer(reserve=N * 4 * 6)
        self.vao = self.ctx.vertex_array(self.prog, [])
        self.vao.transform(self.buffer, vertices=N)

        data = struct.unpack('{}f'.format(N * 6), self.buffer.read())
        for i in range(0, N * 6, 6):
            print(data[i:i + 3], data[i + 3:i + 6])

    def render(self, time, frame_time):
        exit(0)


if __name__ == '__main__':
    GenerateData.run()



import os

import glm
import moderngl
import moderngl_window as mglw


class Example(mglw.WindowConfig):
    title = 'Resource Loading with ModernGL Window'
    resizable = False
    gl_version = (3, 3)
    window_size = (800, 800)
    aspect_ratio = 1.0
    resource_dir = os.path.normpath(os.path.join(__file__, '../data'))

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.program = self.ctx.program(
            vertex_shader='''
                #version 330 core

                vec2 positions[3] = vec2[](
                    vec2(-1.0, -1.0),
                    vec2(3.0, -1.0),
                    vec2(-1.0, 3.0)
                );

                out vec2 v_uv;

                void main() {
                    gl_Position = vec4(positions[gl_VertexID], 0.0, 1.0);
                    v_uv = positions[gl_VertexID];
                }
            ''',
            fragment_shader='''
                #version 330

                in vec2 v_uv;
                out vec4 f_color;

                uniform sampler2D Texture;

                uniform vec2 seed;
                uniform int iter;

                void main() {
                    vec2 c = seed;
                    int i;

                    vec2 z = v_uv * vec2(2.0, 2.0);

                    for (i = 0; i < iter; i++) {
                        float x = (z.x * z.x - z.y * z.y) + c.x;
                        float y = (z.y * z.x + z.x * z.y) + c.y;

                        if ((x * x + y * y) > 4.0) {
                            break;
                        }

                        z.x = x;
                        z.y = y;
                    }

                    f_color = texture(Texture, vec2((i == iter ? 0.0 : float(i)) / 100.0, 0.0));
                }
            ''',
        )

        self.texture = self.load_texture_2d('colormaps/fractal.png')
        self.sampler = self.ctx.sampler(texture=self.texture)
        self.vao = self.ctx.vertex_array(self.program, [])
        self.vao.vertices = 3

    def render(self, time, frame_time):
        self.ctx.clear()
        self.program['seed'] = (-0.8, 0.156)
        self.program['iter'] = 100
        self.sampler.use()
        self.vao.render()


if __name__ == '__main__':
    Example.run()



# Example of creating a window without
# enclosing code in modengl_window.WindowConfig class
# Description: This example shows how to create a window as a separate object.
# Note: Only one window can be used.

import moderngl as gl
import moderngl_window as glw
import struct


# Create window
window_cls = glw.get_local_window_cls('pyglet')
window = window_cls(
    size=(1024, 512), fullscreen=False, title='Separate window',
    resizable=False, vsync=True, gl_version=(3, 3))
ctx = window.ctx
glw.activate_context(window, ctx=ctx)
window.clear()
window.swap_buffers()

# Window prog
prog = ctx.program(
    vertex_shader='''
    #version 330

    in vec3 in_vert;
    in vec4 in_color;

    uniform mat4 prj;

    out vec4 o_color;

    void main() {
        gl_Position = prj*vec4(in_vert.xyz, 1.0);
        o_color = in_color;
    }
    ''',
    fragment_shader='''
    #version 330

    in vec4 o_color;

    out vec4 f_color;

    void main() {
        f_color = o_color;
    }
    ''')

# Preparing
content = [
    .3, .3, -.5, 1, 0, 0, 1,
    .3, -.3, -.5, 1, 0, 0, 1,
    -.3, .3, -.4, 1, 0, 0, 1,

    .3, .4, -.5, 0, 0, 1, 1,
    .3, -.2, -.5, 0, 0, 1, 1,
    -.3, .4, -.45, 0, 0, 1, 1
]
buf = ctx.buffer(struct.pack(f'{len(content)}f', *content))
vao = ctx.vertex_array(prog, buf, 'in_vert', 'in_color')

f = 1
n = 0.1
prj = (
    1, 0, 0, 0,
    0, 1, 0, 0,
    0, 0, -(f+n)/(f-n), -1,
    0, 0, -(2*f*n)/(f-n), 0
)
vao.program['prj'] = prj

# Render loop
ctx.enable(gl.DEPTH_TEST)
while not window.is_closing:
    window.fbo.use()
    window.fbo.clear()
    vao.render()
    window.swap_buffers()



import moderngl
from PIL import Image

ctx = moderngl.create_standalone_context()

fbo = ctx.framebuffer([ctx.renderbuffer((800, 800), components=4, samples=16)])
output = ctx.framebuffer([ctx.renderbuffer((800, 800), components=4)])

prog = ctx.program(
    vertex_shader='''
        #version 330 core

        vec2 vertex[3] = vec2[](
            vec2(1.0, 0.0),
            vec2(-0.5, -0.86),
            vec2(-0.5, 0.86)
        );

        vec3 color[3] = vec3[](
            vec3(0.0, 0.0, 1.0),
            vec3(0.0, 1.0, 0.0),
            vec3(1.0, 0.0, 0.0)
        );

        out vec3 v_color;

        void main() {
            v_color = color[gl_VertexID];
            float r = (gl_InstanceID - 4) * 0.15;
            mat2 rot = mat2(cos(r), sin(r), -sin(r), cos(r));
            gl_Position = vec4(rot * vertex[gl_VertexID] * 0.95, 0.0, 1.0);
        }
    ''',
    fragment_shader='''
        #version 330 core
        in vec3 v_color;
        out vec4 f_color;
        void main() {
            f_color = vec4(pow(v_color, vec3(1.0 / 2.2)), 1.0);
        }
    ''',
)

vao = ctx.vertex_array(prog, [])
vao.instances = 9
vao.vertices = 3

fbo.use()
ctx.clear()
vao.render()

ctx.copy_framebuffer(output, fbo)

img = Image.frombuffer('RGBA', output.size, output.read(components=4)).transpose(Image.Transpose.FLIP_TOP_BOTTOM)
img.save('logo.png')



import math
import os
import struct
import sys

import glm
import moderngl
import pygame
from objloader import Obj
from PIL import Image, ImageDraw, ImageFont

os.environ['SDL_WINDOWS_DPI_AWARENESS'] = 'permonitorv2'

pygame.init()
pygame.display.set_mode((800, 800), flags=pygame.OPENGL | pygame.DOUBLEBUF, vsync=True)


def init_includes():
    ctx = moderngl.get_context()

    ctx.includes['uniform_buffer'] = '''
        struct Light {
            vec4 light_position;
            vec4 light_color;
            float light_power;
        };

        layout (std140) uniform Common {
            mat4 camera;
            vec4 camera_position;
            Light lights[2];
        };
    '''

    ctx.includes['blinn_phong'] = '''
        vec3 blinn_phong(
                vec3 vertex, vec3 normal, vec3 camera_position, vec3 light_position, float shininess, vec3 ambient_color,
                vec3 diffuse_color, vec3 light_color, vec3 spec_color, float light_power) {

            vec3 light_dir = light_position - vertex;
            float light_distance = length(light_dir);
            light_distance = light_distance * light_distance;
            light_dir = normalize(light_dir);

            float lambertian = max(dot(light_dir, normal), 0.0);
            float specular = 0.0;

            if (lambertian > 0.0) {
                vec3 view_dir = normalize(camera_position - vertex);
                vec3 half_dir = normalize(light_dir + view_dir);
                float spec_angle = max(dot(half_dir, normal), 0.0);
                specular = pow(spec_angle, shininess);
            }

            vec3 color_linear = ambient_color +
                diffuse_color * lambertian * light_color * light_power / light_distance +
                spec_color * specular * light_color * light_power / light_distance;

            return color_linear;
        }
    '''

    ctx.includes['calculate_lights'] = '''
        vec3 calculate_lights(vec3 vertex, vec3 normal, vec3 color, vec3 camera_position) {
            vec3 result = vec3(0.0);
            for (int i = 0; i < 2; ++i) {
                result += blinn_phong(
                    vertex, normal, camera_position, lights[i].light_position.xyz, 16.0, color * 0.05,
                    color, lights[i].light_color.rgb, vec3(1.0, 1.0, 1.0), lights[i].light_power
                );
            }
            return result;
        }
    '''

    ctx.includes['srgb'] = '''
        vec3 srgb_to_linear(vec3 color) {
            return pow(color, vec3(2.2));
        }
        vec3 linear_to_srgb(vec3 color) {
            return pow(color, vec3(1.0 / 2.2));
        }
    '''

    ctx.includes['hash13'] = '''
        float hash13(vec3 p3) {
            p3 = fract(p3 * 0.1031);
            p3 += dot(p3, p3.zyx + 31.32);
            return fract((p3.x + p3.y) * p3.z);
        }
    '''


class FilmGrain:
    def __init__(self, texture):
        self.ctx = moderngl.get_context()
        self.program = self.ctx.program(
            vertex_shader='''
                #version 330 core

                vec2 positions[3] = vec2[](
                    vec2(-1.0, -1.0),
                    vec2(3.0, -1.0),
                    vec2(-1.0, 3.0)
                );

                void main() {
                    gl_Position = vec4(positions[gl_VertexID], 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core
                #include "hash13"

                uniform sampler2D Texture;
                uniform float time;

                layout (location = 0) out vec4 out_color;

                void main() {
                    ivec2 at = ivec2(gl_FragCoord.xy);
                    float grain = hash13(vec3(gl_FragCoord.xy, time)) * 0.5 + 0.5;
                    out_color = texelFetch(Texture, at, 0) * grain;
                }
            ''',
        )
        self.sampler = self.ctx.sampler(texture=texture)
        self.vao = self.ctx.vertex_array(self.program, [])
        self.vao.vertices = 3

    def render(self, now):
        self.ctx.enable_only(self.ctx.NOTHING)
        self.sampler.use()
        self.program['time'] = now
        self.vao.render()


class Overlay:
    def __init__(self, size):
        self.ctx = moderngl.get_context()
        self.img = Image.new('RGBA', size)
        self.draw = ImageDraw.Draw(self.img)
        self.draw.font = ImageFont.truetype('examples/data/fonts/OpenSans/OpenSans-Medium.ttf', 20)
        self.texture = self.ctx.texture(size, 4)
        self.program = self.ctx.program(
            vertex_shader='''
                #version 330 core

                vec2 positions[3] = vec2[](
                    vec2(-1.0, -1.0),
                    vec2(3.0, -1.0),
                    vec2(-1.0, 3.0)
                );

                void main() {
                    gl_Position = vec4(positions[gl_VertexID], 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core

                uniform sampler2D Texture;

                layout (location = 0) out vec4 out_color;

                void main() {
                    ivec2 at = ivec2(gl_FragCoord.xy);
                    at.y = textureSize(Texture, 0).y - at.y - 1;
                    out_color = texelFetch(Texture, at, 0);
                }
            ''',
        )
        self.sampler = self.ctx.sampler(texture=self.texture)
        self.vao = self.ctx.vertex_array(self.program, [])
        self.vao.vertices = 3

        self.clock = pygame.time.Clock()
        self.fps = 0.0

    def render(self):
        self.clock.tick()
        mouse = pygame.mouse.get_pos()
        now = pygame.time.get_ticks() / 1000.0
        self.fps = self.fps * 0.95 + self.clock.get_fps() * 0.05

        self.draw.rectangle((0, 0, *self.img.size), fill=(0, 0, 0, 0))
        self.draw.text((100, 100), f'fps: {self.fps:.2f}', fill='#fff')
        self.draw.text((100, 130), f'elapsed: {now:.2f}s', fill='#f00')
        self.draw.text((100, 160), f'mouse: {mouse}', fill='#fff')

        rect = self.draw.textbbox((0, 0), 'Hello, World!')
        pos = mouse[0] - (rect[0] + rect[2]) / 2, mouse[1] - 40
        self.draw.text(pos, 'Hello, World!', fill='#fff')

        self.texture.write(self.img.tobytes())
        self.ctx.enable_only(self.ctx.BLEND)
        self.sampler.use()
        self.vao.render()


class UniformBuffer:
    def __init__(self):
        self.ctx = moderngl.get_context()
        self.data = bytearray(1024)
        self.ubo = self.ctx.buffer(self.data)

    def set_camera(self, eye, target):
        proj = glm.perspective(45.0, 1.0, 0.1, 1000.0)
        look = glm.lookAt(eye, target, (0.0, 0.0, 1.0))
        camera = proj * look
        self.data[0:64] = camera.to_bytes()
        self.data[64:80] = struct.pack('4f', *eye, 0.0)

    def set_light(self, light_index, position, color, power):
        offset = 80 + light_index * 48
        self.data[offset + 0 : offset + 16] = struct.pack('4f', *position, 0.0)
        self.data[offset + 16 : offset + 32] = struct.pack('4f', *color, 0.0)
        self.data[offset + 32 : offset + 36] = struct.pack('f', power)

    def use(self):
        self.ubo.write(self.data)
        self.ubo.bind_to_uniform_block()


class ColorMaterial:
    def __init__(self):
        self.ctx = moderngl.get_context()
        self.program = self.ctx.program(
            vertex_shader='''
                #version 330 core
                #include "uniform_buffer"

                uniform vec3 position;
                uniform float scale;

                layout (location = 0) in vec3 in_vertex;
                layout (location = 1) in vec3 in_normal;
                layout (location = 2) in vec2 in_uv;

                out vec3 v_vertex;
                out vec3 v_normal;
                out vec2 v_uv;

                void main() {
                    v_vertex = position + in_vertex * scale;
                    v_normal = in_normal;
                    v_uv = in_uv;

                    gl_Position = camera * vec4(v_vertex, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core
                #include "uniform_buffer"
                #include "blinn_phong"
                #include "calculate_lights"
                #include "srgb"

                uniform vec3 color;

                in vec3 v_vertex;
                in vec3 v_normal;
                in vec2 v_uv;

                layout (location = 0) out vec4 out_color;

                void main() {
                    vec3 color_linear = calculate_lights(v_vertex, v_normal, srgb_to_linear(color), camera_position.xyz);
                    out_color = vec4(linear_to_srgb(color_linear), 1.0);
                }
            ''',
        )

        self.color = (1.0, 1.0, 1.0)

    def use(self):
        self.program['color'] = self.color

    def vertex_array(self, buffer):
        return self.ctx.vertex_array(self.program, [(buffer, '3f 3f 8x', 'in_vertex', 'in_normal')])


class TextureMaterial:
    def __init__(self, texture):
        self.texture = texture
        self.ctx = moderngl.get_context()
        self.program = self.ctx.program(
            vertex_shader='''
                #version 330 core
                #include "uniform_buffer"

                uniform vec3 position;
                uniform float scale;

                layout (location = 0) in vec3 in_vertex;
                layout (location = 1) in vec3 in_normal;
                layout (location = 2) in vec2 in_uv;

                out vec3 v_vertex;
                out vec3 v_normal;
                out vec2 v_uv;

                void main() {
                    v_vertex = position + in_vertex * scale;
                    v_normal = in_normal;
                    v_uv = in_uv;

                    gl_Position = camera * vec4(v_vertex, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core
                #include "uniform_buffer"
                #include "blinn_phong"
                #include "calculate_lights"
                #include "srgb"

                uniform sampler2D Texture;

                in vec3 v_vertex;
                in vec3 v_normal;
                in vec2 v_uv;

                layout (location = 0) out vec4 out_color;

                void main() {
                    vec3 color = texture(Texture, v_uv).rgb;
                    vec3 color_linear = calculate_lights(v_vertex, v_normal, srgb_to_linear(color), camera_position.xyz);
                    out_color = vec4(linear_to_srgb(color_linear), 1.0);
                }
            ''',
        )

    def use(self):
        self.texture.use()

    def vertex_array(self, buffer):
        return self.ctx.vertex_array(self.program, [(buffer, '3f 3f 2f', 'in_vertex', 'in_normal', 'in_uv')])


class ImageTexture:
    def __init__(self, path):
        self.ctx = moderngl.get_context()

        img = Image.open(path).convert('RGBA')
        self.texture = self.ctx.texture(img.size, 4, img.tobytes())
        self.sampler = self.ctx.sampler(texture=self.texture)

    def use(self):
        self.sampler.use()


class ModelGeometry:
    def __init__(self, path):
        self.ctx = moderngl.get_context()

        obj = Obj.open(path)
        self.vbo = self.ctx.buffer(obj.pack('vx vy vz nx ny nz tx ty'))


class Mesh:
    def __init__(self, material, geometry):
        self.ctx = moderngl.get_context()
        self.vao = material.vertex_array(geometry.vbo)
        self.material = material

    def render(self, position, scale):
        self.material.use()
        self.vao.program['position'] = position
        self.vao.program['scale'] = scale
        self.vao.render()


class Scene:
    def __init__(self):
        self.ctx = moderngl.get_context()

        size = pygame.display.get_window_size()
        self.screen = self.ctx.texture(size, 4)
        self.depth = self.ctx.depth_texture(size)

        self.framebuffer = self.ctx.framebuffer(
            color_attachments=[self.screen],
            depth_attachment=self.depth,
        )

        self.uniform_buffer = UniformBuffer()
        self.texture = ImageTexture('examples/data/textures/crate.png')

        self.color_material = ColorMaterial()
        self.texture_material = TextureMaterial(self.texture)

        self.car_geometry = ModelGeometry('examples/data/models/lowpoly_toy_car.obj')
        self.car = Mesh(self.color_material, self.car_geometry)

        self.crate_geometry = ModelGeometry('examples/data/models/crate.obj')
        self.crate = Mesh(self.texture_material, self.crate_geometry)

        self.film_grain = FilmGrain(self.screen)
        self.overlay = Overlay(size)


    def render(self):
        now = pygame.time.get_ticks() / 1000.0

        eye = (math.cos(now), math.sin(now), 0.5)

        self.framebuffer.use()

        self.ctx.clear()
        self.ctx.enable(self.ctx.DEPTH_TEST)

        self.uniform_buffer.set_camera(eye, (0.0, 0.0, 0.0))
        self.uniform_buffer.set_light(
            light_index=0,
            position=(1.0, 2.0, 3.0),
            color=(1.0, 1.0, 1.0),
            power=10.0,
        )
        self.uniform_buffer.set_light(
            light_index=1,
            position=(-2.0, -1.0, 4.0),
            color=(1.0, 1.0, 1.0),
            power=10.0,
        )
        self.uniform_buffer.use()

        self.crate.render((0.0, 0.0, 0.0), 0.2)

        self.color_material.color = (1.0, 0.0, 0.0)
        self.car.render((-0.4, 0.0, 0.0), 0.2)

        self.color_material.color = (0.0, 0.0, 1.0)
        self.car.render((0.4, 0.0, 0.0), 0.2)

        self.ctx.screen.use()
        self.film_grain.render(now)
        self.overlay.render()


init_includes()
scene = Scene()

while True:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            pygame.quit()
            sys.exit()

    scene.render()

    pygame.display.flip()



# WebGL Water
# https://madebyevan.com/webgl-water/
# Copyright 2011 Evan Wallace
# Released under the MIT license

import re
import moderngl
from typing import Self
from PIL import Image
import numpy as np
import glm
from typing import Iterator


class Matrices:
    def __init__(self, model: glm.mat4 = None, view: glm.mat4 = None, projection: glm.mat4 = None) -> None:
        self._model = model or glm.identity(glm.mat4)
        self.view = view or glm.lookAt(
            (1.0, 1.0, 1.0), (0.0, 0.0, 0.0), (0.0, 1.0, 0.0))
        self.projection = projection or glm.perspective(45.0, 1.0, 0.1, 1000.0)

    @property
    def normal(self) -> glm.mat4:
        return glm.transpose(glm.inverse(self._model))

    @property
    def model_view(self) -> glm.mat4:
        return self.view * self._model

    @property
    def model_view_proj(self) -> glm.mat4:
        return self.projection * self.model_view

    @property
    def eye(self) -> glm.vec3:
        m = np.array(self.model_view).T
        axis_x: np.ndarray = m[0]
        axis_y: np.ndarray = m[1]
        axis_z: np.ndarray = m[2]
        offset: np.ndarray = m[3]
        return glm.vec3(
            -offset.dot(axis_x), -offset.dot(axis_y), -offset.dot(axis_z))

    def get(self, cap_name: str) -> glm.mat4:
        match cap_name:
            case "MVM":
                return self.model_view
            case "MVMI":
                return glm.inverse(self.model_view)
            case "PM":
                return self.projection
            case "PMI":
                return glm.inverse(self.projection)
            case "MVPM":
                return self.model_view_proj
            case "MVPMI":
                return glm.inverse(self.model_view_proj)
            case "NM":
                return self.normal


class MeshData:
    def __init__(self, vertices: np.ndarray, triangles: np.ndarray, coords: np.ndarray = None, normals: np.ndarray = None, color: np.ndarray = None, lines: np.ndarray = None) -> None:
        self.vertices = vertices
        self.triangles = triangles
        self.coords = coords
        self.normals = normals
        self.color = color
        self.lines = lines


class Mesh:
    def __init__(self, data: MeshData, ctx: moderngl.Context = None) -> None:
        ctx = ctx or moderngl.get_context()
        self.vbos = {
            "gl_Vertex": Mesh.layout(ctx, data.vertices, '3f'),
            "gl_TexCoord": Mesh.layout(ctx, data.coords, '2f'),
            "gl_Normal": Mesh.layout(ctx, data.normals, '3f'),
            "gl_Color": Mesh.layout(ctx, data.color, '4f'),
        }
        self.lines_ibo = ctx.buffer(data.lines.astype("i4").tobytes())
        if data.triangles is not None:
            self.triangles_ibo = ctx.buffer(
                data.triangles.astype("i4").tobytes())
        self.vaos: dict[int, tuple[moderngl.VertexArray, int]] = {}

    def layout(ctx: moderngl.Context, array: np.ndarray, t: str) -> bytes:
        if array is not None:
            return (ctx.buffer(array.astype("f4").tobytes()), t)
        return None

    def _layouts(self, prefix: str, used_attributes: set[str]) -> Iterator[tuple]:
        for attr_name in self.vbos:
            if self.vbos[attr_name] is None:
                continue
            prefixed_name = prefix + attr_name
            if prefixed_name in used_attributes:
                yield (*self.vbos[attr_name], prefixed_name)

    def build_vao(self, ctx: moderngl.Context, program: moderngl.Program, layouts: list, mode=moderngl.TRIANGLES):
        if mode == moderngl.TRIANGLES:
            vao = ctx.vertex_array(
                program, layouts, self.triangles_ibo, mode=moderngl.TRIANGLES)
        else:
            vao = ctx.vertex_array(
                program, layouts, self.lines_ibo, mode=moderngl.LINES)
        self.vaos[program.glo] = (vao, mode)
        return vao

    def vertex_array(self, ctx: moderngl.Context, program: moderngl.Program, prefix: str, used_attributes: dict[str, moderngl.Attribute], mode=moderngl.TRIANGLES) -> moderngl.VertexArray:
        vao, vao_mode = self.vaos.get(program.glo, (None, None))
        if vao is not None:
            if vao_mode == mode:
                return vao
            else:
                vao.release()

        layouts = list(self._layouts(prefix, used_attributes))
        vao = self.build_vao(ctx, program, layouts, mode)
        return vao


class MeshBuilder(MeshData):
    def __init__(self, vertices: np.ndarray, triangles: np.ndarray, coords: np.ndarray = None, normals: np.ndarray = None, lines: np.ndarray = None) -> None:
        super().__init__(vertices, triangles, coords, normals, None, lines)
        if self.normals is None:
            self.compute_normals()
        if self.lines is None:
            self.compute_wireframe()

    def compute_normals(self):
        self.normals = np.zeros(np.shape(self.vertices))
        pass

    def compute_wireframe(self):
        lines = {}
        for i in range(len(self.triangles)):
            t = self.triangles[i]
            for j in range(3):
                a = t[j]
                b = t[(j+1) % 3]
                l = (min(a, b), max(a, b))
                k = (l[0] << 16) + l[1]
                lines[k] = l
        lines = list(lines.values())
        self.lines = np.array(lines, dtype=np.int32)

    def build(self, ctx: moderngl.Context = None) -> Mesh:
        return Mesh(self, ctx)

    def seg_lines(*points: glm.vec3):
        vertices = list(points)
        num_vertices = len(vertices)
        num_vertices = num_vertices - (num_vertices % 2)
        vertices = np.array(vertices[0:num_vertices])
        lines = np.arange(num_vertices, dtype=np.int32)
        return MeshBuilder(vertices, None, normals=np.zeros(vertices.shape), lines=lines)

    def panel(detail: int = 1, detail_x: int = None, detail_y: int = None) -> Self:
        detail_x = detail_x or detail
        detail_y = detail_y or detail
        vertices = []
        coords = []
        normals = []
        triangles = []
        for y in range(detail_y + 1):
            t = y / detail_y
            for x in range(detail_x + 1):
                s = x / detail_x
                vertices.append((2.0 * s - 1.0, 2.0 * t - 1.0, 0.0))
                coords.append((s, t))
                normals.append((0.0, 0.0, 1.0))
                if x < detail_x and y < detail_y:
                    i = x + y * (detail_y + 1)
                    triangles.append((i, i + 1, i + detail_x + 1))
                    triangles.append(
                        (i + detail_x + 1, i + 1, i + detail_x + 2))
        return MeshBuilder(np.array(vertices), np.array(triangles, dtype=np.int32), np.array(coords), np.array(normals))

    cube_data = [
        [0, 4, 2, 6, -1, 0, 0],
        [1, 3, 5, 7, +1, 0, 0],
        [0, 1, 4, 5, 0, -1, 0],
        [2, 6, 3, 7, 0, +1, 0],
        [0, 2, 1, 3, 0, 0, -1],
        [4, 5, 6, 7, 0, 0, +1]
    ]

    def _pick_octant(i) -> tuple[int, int, int]:
        return ((i & 1) * 2 - 1, (i & 2) - 1, (i & 4) / 2 - 1)

    def cube() -> Self:
        vertices = []
        coords = []
        normals = []
        triangles = []
        for i in range(len(MeshBuilder.cube_data)):
            data = MeshBuilder.cube_data[i]
            v = i * 4
            for j in range(4):
                d = data[j]
                vertices.append(MeshBuilder._pick_octant(d))
                coords.append((j & 1, (j & 2) / 2))
                normals.append((data[4:7]))
            triangles.append((v, v+1, v+2))
            triangles.append((v+2, v+1, v+3))
        return MeshBuilder(np.array(vertices), np.array(triangles, dtype=np.int32), np.array(coords), np.array(normals))

    def _tri(flip: bool, a, b, c):
        return (a, c, b) if flip else (a, b, c)

    def _fix(x: np.ndarray) -> np.ndarray:
        return x + (x - x * x) / 2.0

    def _tri_idx(d: int, i: int, j: int) -> int:
        return i * d + (i - i * i) // 2 + j

    def sphere(detail: int = 6):
        detail_2 = detail * 2
        detail_22 = detail_2 * detail_2
        raw_vertices = []
        raw_abc = []
        vertices = []
        coords = []
        unique_map = {}
        for i in range(detail + 1):
            for j in range(detail - i + 1):
                raw_abc.append((i, j, detail - i - j))
        raw_vertices = MeshBuilder._fix((np.array(raw_abc) / detail))
        raw_normal = np.linalg.norm(raw_vertices, axis=1)
        raw_vertices = raw_vertices / raw_normal[:, np.newaxis]

        triangles = []
        for octant in range(8):
            scale = MeshBuilder._pick_octant(octant)
            flip = scale[0] * scale[1] * scale[2]
            data = []
            for abc, rv in zip(raw_abc, raw_vertices):
                a, b, c = abc
                k = a * scale[0] * detail_22 + b * \
                    scale[1] * detail_2 + c * scale[2]
                if k not in unique_map:
                    unique_map[k] = len(vertices)
                    vertex = rv * np.array(scale)
                    coord = (1 - a, c) if scale[1] > 0 else (c, 1 - a)
                    vertices.append(vertex)
                    coords.append(coord)
                data.append(unique_map[k])

            for i in range(1, detail + 1):
                for j in range(detail - i + 1):
                    a = MeshBuilder._tri_idx(detail + 1, i-1, j)
                    b = MeshBuilder._tri_idx(detail + 1, i, j)
                    triangles.append(MeshBuilder._tri(
                        flip, data[a], data[a + 1], data[b]))
                    if (i + j < detail):
                        triangles.append(MeshBuilder._tri(
                            flip, data[b], data[a + 1], data[b+1]))

        return MeshBuilder(np.array(vertices), np.array(triangles, dtype=np.int32), np.array(coords))


class Shader:
    header = """
        uniform mat3 gl_NormalMatrix;
        uniform mat4 gl_ModelViewMatrix;
        uniform mat4 gl_ProjectionMatrix;
        uniform mat4 gl_ModelViewProjectionMatrix;
        uniform mat4 gl_ModelViewMatrixInverse;
        uniform mat4 gl_ProjectionMatrixInverse;
        uniform mat4 gl_ModelViewProjectionMatrixInverse;
    """

    _vert_header = """
        attribute vec4 gl_Vertex;
        attribute vec2 gl_TexCoord;
        attribute vec3 gl_Normal;
        attribute vec4 gl_Color;
        vec4 ftransform() {
            return gl_ModelViewProjectionMatrix * gl_Vertex;
        }
    """
    vert_header = "#version 130\n" + header + _vert_header

    frag_header = """
        #version 130
        precision highp float;
    """ + header

    matrices_name = list(
        map(lambda m: m.groups()[0], re.finditer(r"\b(gl_[^;]*)\b;", header)))
    attributes_name = list(
        map(lambda m: m.groups()[0], re.finditer(r"\b(gl_[^;]*)\b;", _vert_header)))

    def dump_source(emsg: str, vert_shader: str, frag_shader: str):
        if emsg.find("vertex_shader") != -1:
            lines = vert_shader.splitlines()
            for num, line in enumerate(lines):
                print(f"{num+1:4d}", line)
        if emsg.find("fragment_shader") != -1:
            lines = frag_shader.splitlines()
            for num, line in enumerate(lines):
                print(f"{num+1:4d}", line)

    def fix(header: str, source: str, prefix: str) -> str:
        replace = {}
        source = header + re.sub(r"#extension.*\n", '', source)
        all_gl_vars = list(
            map(lambda m: m.groups()[0], re.finditer(r"\b(gl_\w+)\b;", header)))
        for gl_var_name in all_gl_vars:
            if gl_var_name not in replace:
                source = re.sub('\\b' + gl_var_name + '\\b',
                                prefix + gl_var_name, source)
        return source

    def __init__(self, vert_shader: str, frag_shader: str, ctx: moderngl.Context = None, prefix="LIGHTGL") -> None:
        self.ctx = ctx or moderngl.get_context()
        source: str = vert_shader + frag_shader
        self.use_matrices: dict[str, str] = {}
        for name in Shader.matrices_name:
            if source.find(name) != -1:
                # gl_NormalMatrix -> NM
                capitalLetters = re.sub(r"[a-z_]", '', name)
                self.use_matrices[capitalLetters] = prefix + name

        if source.find('ftransform') != -1:
            self.use_matrices['MVPM'] = prefix + "gl_ModelViewProjectionMatrix"

        self.prefix = prefix
        vertex_shader = Shader.fix(Shader.vert_header, vert_shader, prefix)
        fragment_shader = Shader.fix(Shader.frag_header, frag_shader, prefix)
        try:
            self.program = self.ctx.program(vertex_shader, fragment_shader)
        except moderngl.Error as e:
            Shader.dump_source(e.args[0], vertex_shader, fragment_shader)
            print(e)
            exit(-1)
        self.use_attributes = {name: self.program[name] for name in self.program
                               if isinstance(self.program[name], moderngl.Attribute)}

    def draw_mesh(self, mesh: Mesh, matrices: Matrices = None, unifroms: dict = {}, mode=moderngl.TRIANGLES):
        if matrices is not None:
            for m_cap_name in self.use_matrices:
                self.program[self.use_matrices[m_cap_name]].write(
                    matrices.get(m_cap_name))
        for k in unifroms:
            self.program[k] = unifroms[k]
        vao = mesh.vertex_array(self.ctx, self.program,
                                self.prefix, self.use_attributes, mode)
        vao.render(mode)


class Texture:
    def __init__(self, texture: moderngl.Texture = None, sampler: moderngl.Sampler = None) -> None:
        self.texture: moderngl.Texture = texture
        self.sampler: moderngl.Sampler = sampler
        self.framebuffer: moderngl.Framebuffer = None

    def use(self, location=0):
        if self.sampler:
            self.sampler.use(location)
        else:
            self.texture.use(location)

    def draw_to(self, ctx: moderngl.Context = None):
        if self.framebuffer is None:
            ctx = ctx or moderngl.get_context()
            self.framebuffer = ctx.framebuffer(
                self.texture, ctx.depth_renderbuffer(self.texture.size))
        self.framebuffer.use()

    def swap_with(self, other: Self):
        tmp_texture = self.texture
        tmp_sampler = self.sampler
        tmp_framebuffer = self.framebuffer
        self.texture = other.texture
        self.sampler = other.sampler
        self.framebuffer = other.framebuffer
        other.texture = tmp_texture
        other.sampler = tmp_sampler
        other.framebuffer = tmp_framebuffer


class RawTexture(Texture):
    def __init__(self, size: tuple[int, int], data: bytes = None, pixel_fmt='RGBA', ctx: moderngl.Context = None, repeat=(False, False, False), filter=(moderngl.LINEAR, moderngl.LINEAR), mipmaps=False, dtype="f1") -> None:
        ctx = ctx or moderngl.get_context()
        texture = ctx.texture(size, len(pixel_fmt), data, dtype=dtype)
        sampler = ctx.sampler(
            repeat_x=repeat[0], repeat_y=repeat[1], repeat_z=repeat[2], filter=filter, texture=texture)
        if mipmaps:
            texture.build_mipmaps()
        super().__init__(texture, sampler)


class ImageTexture(RawTexture):
    def __init__(self, image_path: str, pixel_fmt='RGBA', ctx: moderngl.Context = None, repeat=(False, False, False), filter=(moderngl.LINEAR, moderngl.LINEAR), mipmaps=False) -> None:
        img = Image.open(image_path).transpose(
            Image.FLIP_TOP_BOTTOM).convert(pixel_fmt)
        super().__init__(img.size, img.tobytes(),
                         pixel_fmt, ctx, repeat, filter, mipmaps)


def Cubemap(pos_x: str, neg_x: str, pos_y: str, neg_y: str, pos_z: str, neg_z: str, ctx: moderngl.Context = None) -> moderngl.TextureCube:
    ctx = ctx or moderngl.get_context()
    images_path = (pos_x, neg_x, pos_y, neg_y, pos_z, neg_z)
    images: list[Image.Image] = list(map(lambda p: Image.open(p).transpose(
        Image.FLIP_TOP_BOTTOM).convert("RGB"), images_path))
    image_size = images[0].size
    image_len = image_size[0] * image_size[1] * 3
    data = bytearray(image_len * 6)
    for i in range(6):
        data[i * image_len:] = images[i].tobytes()
    return ctx.texture_cube(image_size, 3, data)



import moderngl
from PIL import Image

ctx = moderngl.create_standalone_context()

fbo1 = ctx.framebuffer([ctx.renderbuffer((800, 800), components=4, samples=16)])
fbo2 = ctx.framebuffer([ctx.renderbuffer((800, 800), components=4)])

prog = ctx.program(
    vertex_shader=open('examples/data/logo.vert.spv', 'rb').read(),
    fragment_shader=open('examples/data/logo.frag.spv', 'rb').read(),
)

vao = ctx.vertex_array(prog, [])
vao.scope = ctx.scope(fbo1)
vao.instances = 9
vao.vertices = 3

fbo1.clear(0.0, 0.0, 0.0, 0.0)
vao.render()

ctx.copy_framebuffer(fbo2, fbo1)

img = Image.frombuffer('RGBA', fbo2.size, fbo2.read(components=4)).transpose(Image.Transpose.FLIP_TOP_BOTTOM)
img.show()



import math
import os
import sys

import glm
import moderngl
import numpy as np
import pygame

os.environ['SDL_WINDOWS_DPI_AWARENESS'] = 'permonitorv2'

pygame.init()
pygame.display.set_mode((800, 800), flags=pygame.OPENGL | pygame.DOUBLEBUF, vsync=True)


class TriangleGeometry:
    def __init__(self):
        self.ctx = moderngl.get_context()
        vertices = np.array([
            0.0, 0.4, 0.0,
            -0.4, -0.3, 0.0,
            0.4, -0.3, 0.0,
        ])

        self.vbo = self.ctx.buffer(vertices.astype('f4').tobytes())

    def vertex_array(self, program):
        return self.ctx.vertex_array(program, [(self.vbo, '3f', 'in_vertex')])


class PlaneGeometry:
    def __init__(self):
        self.ctx = moderngl.get_context()
        vertices = np.array([
            -0.5, -0.5, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0,
            0.5, -0.5, 0.0, 0.0, 0.0, 1.0, 1.0, 0.0,
            -0.5, 0.5, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0,
            -0.5, 0.5, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0,
            0.5, -0.5, 0.0, 0.0, 0.0, 1.0, 1.0, 0.0,
            0.5, 0.5, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0,
        ])

        self.vbo = self.ctx.buffer(vertices.astype('f4').tobytes())

    def vertex_array(self, program):
        return self.ctx.vertex_array(program, [(self.vbo, '3f 12x 8x', 'in_vertex')])


class Mesh:
    def __init__(self, program, geometry):
        self.ctx = moderngl.get_context()
        self.vao = geometry.vertex_array(program)

    def render(self, position, color, scale):
        self.vao.program['position'] = position
        self.vao.program['color'] = color
        self.vao.program['scale'] = scale
        self.vao.render()


class Scene:
    def __init__(self):
        self.ctx = moderngl.get_context()

        self.program = self.ctx.program(
            vertex_shader='''
                #version 330 core

                uniform mat4 camera;
                uniform vec3 position;
                uniform float scale;

                layout (location = 0) in vec3 in_vertex;
                layout (location = 1) in vec3 in_normal;
                layout (location = 2) in vec2 in_uv;

                out vec3 v_vertex;
                out vec3 v_normal;
                out vec2 v_uv;

                void main() {
                    v_vertex = position + in_vertex * scale;
                    v_normal = in_normal;
                    v_uv = in_uv;

                    gl_Position = camera * vec4(v_vertex, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core

                uniform vec3 color;

                in vec3 v_vertex;
                in vec3 v_normal;
                in vec2 v_uv;

                layout (location = 0) out vec4 out_color;

                void main() {
                    out_color = vec4(color, 1.0);
                }
            ''',
        )

        self.triangle_geometry = TriangleGeometry()
        self.triangle = Mesh(self.program, self.triangle_geometry)

        self.plane_geometry = PlaneGeometry()
        self.plane = Mesh(self.program, self.plane_geometry)

    def camera_matrix(self):
        now = pygame.time.get_ticks() / 1000.0
        eye = (math.cos(now), math.sin(now), 0.5)
        proj = glm.perspective(45.0, 1.0, 0.1, 1000.0)
        look = glm.lookAt(eye, (0.0, 0.0, 0.0), (0.0, 0.0, 1.0))
        return proj * look

    def render(self):
        camera = self.camera_matrix()

        self.ctx.clear()
        self.ctx.enable(self.ctx.DEPTH_TEST)

        self.program['camera'].write(camera)

        self.triangle.render((-0.2, 0.0, 0.0), (1.0, 0.0, 0.0), 0.2)
        self.plane.render((0.0, 0.0, 0.0), (0.0, 1.0, 0.0), 0.2)
        self.triangle.render((0.2, 0.0, 0.0), (0.0, 0.0, 1.0), 0.2)


scene = Scene()

while True:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            pygame.quit()
            sys.exit()

    scene.render()

    pygame.display.flip()



import pygame
import moderngl

pygame.init()
pygame.display.set_mode((400, 400), flags=pygame.OPENGL | pygame.DOUBLEBUF)

# ctx = moderngl.create_context()
ctx = moderngl.get_context()

prog = ctx.program(
    vertex_shader='''
        #version 330 core
        uniform float time;
        vec2 vertex[3] = vec2[](
            vec2(1.0, 0.0),
            vec2(-0.5, -0.86),
            vec2(-0.5, 0.86)
        );
        vec3 color[3] = vec3[](
            vec3(0.0, 0.0, 1.0),
            vec3(0.0, 1.0, 0.0),
            vec3(1.0, 0.0, 0.0)
        );
        out vec3 v_color;
        void main() {
            v_color = color[gl_VertexID];
            float r = float(gl_InstanceID - 4) * 0.15 * (0.8 + sin(time) * 0.2);
            mat2 rot = mat2(cos(r), sin(r), -sin(r), cos(r));
            gl_Position = vec4(rot * vertex[gl_VertexID] * 0.95, 0.0, 1.0);
        }
    ''',
    fragment_shader='''
        #version 330 core
        in vec3 v_color;
        out vec4 f_color;
        void main() {
            f_color = vec4(pow(v_color, vec3(1.0 / 2.2)), 1.0);
        }
    ''',
)

vao = ctx.vertex_array(prog, [])
vao.instances = 9
vao.vertices = 3

clock = pygame.Clock()

while True:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            pygame.quit()
            quit()

    prog["time"].value += 1.0 / 60.0
    ctx.clear()
    vao.render()

    pygame.display.flip()
    clock.tick(60)
